// SPDX-License-Identifier: GPL-2.0 OR MIT

//! KMS driver abstractions for rust.

pub mod atomic;
pub mod blend;
pub mod connector;
pub mod crtc;
pub mod encoder;
pub mod framebuffer;
pub mod mode;
pub mod plane;
pub mod gem_atomic_helper;

use crate::{
    drm::{
        drv,
        device::{Device, PendingDevice}
    },
    device,
    prelude::*,
    types::*,
    error::to_result,
    private::Sealed
};
use core::{
    ops::Deref,
    ptr::{self, NonNull},
    mem::{self, ManuallyDrop},
};
use bindings;

#[derive(Copy, Clone)]
pub struct ModeConfigInfo {
    /// The minimum (w, h) resolution this driver can support
    pub min_resolution: (i32, i32),
    /// The maximum (w, h) resolution this driver can support
    pub max_resolution: (i32, i32),
    /// The maximum (w, h) cursor size this driver can support
    pub max_cursor: (u32, u32),
    /// The preferred depth for dumb ioctls
    pub preferred_depth: u32,
}

/// An atomic KMS driver implementation
pub trait KmsDriver: drv::Driver { }

impl<T: KmsDriver> Device<T> {
    const KMS_VTABLE: bindings::drm_mode_config_funcs = bindings::drm_mode_config_funcs {
        atomic_check: Some(bindings::drm_atomic_helper_check),
        // TODO TODO: There are other possibilities then this function, but we need
        // to write up more bindings before we can support those
        fb_create: Some(bindings::drm_gem_fb_create),
        mode_valid: None, // TODO
        atomic_commit: Some(bindings::drm_atomic_helper_commit),
        get_format_info: None,
        atomic_state_free: None,
        atomic_state_alloc: None,
        atomic_state_clear: None,
        output_poll_changed: None,
    };

    const KMS_HELPER_VTABLE: bindings::drm_mode_config_helper_funcs =
        bindings::drm_mode_config_helper_funcs {
            atomic_commit_setup: None, // TODO
            atomic_commit_tail: None, // TODO
        };

    /// Acquire the mode config lock for this `Device<T>`
    #[inline]
    pub fn mode_config_lock(&self) -> ModeConfigGuard<'_, T> {
        // SAFETY: This lock is initialized for as long as `Device<T>` is exposed to users
        unsafe { bindings::mutex_lock(&mut (*self.as_raw()).mode_config.mutex) };

        ModeConfigGuard { dev: self }
    }

    /// Call the reset hooks for all registered modesetting objects
    ///
    /// Drivers should call this before registering a new KMS device if they do not read in hardware
    /// state during init.
    #[inline]
    pub fn mode_config_reset(&self) {
        unsafe { bindings::drm_mode_config_reset(self.as_raw()) }
    }

    /// Setup a generic fbdev device
    #[inline]
    pub fn fbdev_generic_setup(&self) {
        let drm = self.as_raw();

        // SAFETY: The previous build assertion ensures this can only be called for devices with KMS
        // support, which means mode_config is initialized
        unsafe { bindings::drm_fbdev_generic_setup(drm, (*drm).mode_config.preferred_depth) }
    }

    /// Return the number of registered planes
    #[inline]
    pub fn num_plane(&self) -> i32 {
        // SAFETY: The only context which this could change is before registration, which must be
        // single-threaded anyway - so it's safe to just read this value
        unsafe { (*self.as_raw()).mode_config.num_total_plane }
    }
}

impl<T: KmsDriver> PendingDevice<T> {
    pub fn setup_kms(&self, mode_config_info: &ModeConfigInfo) -> Result {
        build_assert!((<T as drv::Driver>::FEATURES & (drv::FEAT_MODESET | drv::FEAT_ATOMIC)) != 0);

        let dev: &device::Device = self.as_ref();

        if self.kms_initialized() {
            dev.pr_err(format_args!("setup_kms() can only be called once\n"));
            return Err(EINVAL);
        }

        // SAFETY: Our interface is proof that this is all taking place before device registration,
        // which means that this is safe
        unsafe {
            (*self.as_raw()).mode_config = bindings::drm_mode_config {
                funcs: &Device::<T>::KMS_VTABLE,
                helper_private: &Device::<T>::KMS_HELPER_VTABLE,
                min_width: mode_config_info.min_resolution.0,
                min_height: mode_config_info.min_resolution.1,
                max_width: mode_config_info.max_resolution.0,
                max_height: mode_config_info.max_resolution.1,
                cursor_width: mode_config_info.max_cursor.0,
                cursor_height: mode_config_info.max_cursor.1,
                preferred_depth: mode_config_info.preferred_depth,
                ..Default::default()
            }
        };

        // SAFETY: We just setup all of the required info this function needs in `drm_device`
        unsafe { to_result(bindings::drmm_mode_config_init(self.as_raw())) }
    }
}

/// Main trait for a modesetting object in DRM
pub trait ModeObject: Sealed + Send + Sync {
    /// The parent driver for this ModeObject
    type Driver: KmsDriver;

    /// Return the `drv::Device` for this `ModeObject`
    fn drm_dev(&self) -> &Device<Self::Driver>;

    /// Return a pointer to the `bindings::drm_mode_object` for this `ModeObject`
    fn raw_mode_obj(&self) -> *mut bindings::drm_mode_object;
}

/// A trait for modesetting objects which don't come with their own reference-counting.
///
/// Objects without a reference count share the lifetime of their parent DRM device
///
/// SAFETY: This trait must not be implemented for modesetting objects which have a refcount
/// already, as otherwise `KmsRef` can't safely guarantee the object will stay alive.
pub unsafe trait StaticModeObject: ModeObject {}

/// An owned reference to a non-reference counted modesetting object.
///
/// In KMS: some modesetting objects aren't reference counted and instead share the drm device's
/// lifetime. This type is order to allow rust drivers access to "owned" references to such objects,
/// and does so by owning a reference to the DRM device which owns the mode object - thus
/// guaranteeing it remains alive for as long as this type is held..
pub struct KmsRef<T: StaticModeObject> {
    object: NonNull<T>,
}

// SAFETY: Owned references to DRM device are thread-safe, and object will survive as long as we
// have said owned references
unsafe impl<T: StaticModeObject> Send for KmsRef<T> {}
unsafe impl<T: StaticModeObject> Sync for KmsRef<T> {}

impl<T: StaticModeObject> From<&T> for KmsRef<T> {
    fn from(value: &T) -> Self {
        // We will drop the reference we leak here in Drop
        value.drm_dev().inc_ref();

        Self { object: value.into() }
    }
}

impl<T: StaticModeObject> Drop for KmsRef<T> {
    fn drop(&mut self) {
        // SAFETY: We're reclaiming the reference we leaked in From<&T>
        drop(unsafe { ARef::from_raw(self.object.as_ref().drm_dev().into()) })
    }
}

impl<T: StaticModeObject> Deref for KmsRef<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        // SAFETY: We're guaranteed object will point to a valid object as long as we hold dev
        unsafe { self.object.as_ref() }
    }
}

impl<T: StaticModeObject> Clone for KmsRef<T> {
    fn clone(&self) -> Self {
        self.drm_dev().inc_ref();

        Self {
            object: self.object
        }
    }
}

/// A trait for mode objects which are ref-counted.
///
/// SAFETY: This trait must not be implemented for modesetting objects which do not have a refcount.
pub unsafe trait RcModeObject: ModeObject {}

unsafe impl<T: RcModeObject> AlwaysRefCounted for T {
    fn inc_ref(&self) {
        // SAFETY: FFI call with no special requirements
        unsafe { bindings::drm_mode_object_get(self.raw_mode_obj()) }
    }

    unsafe fn dec_ref(obj: ptr::NonNull<Self>) {
        // SAFETY: We never expose modesetting objects in our interfaces to users before they're
        // initialized
        unsafe { bindings::drm_mode_object_put(obj.as_ref().raw_mode_obj()) }
    }
}

/// A mode config guard.
///
/// This is an exclusive primitive that represents when `bindings::drm_device.mode_config.lock` is
/// held - as some modesetting operations (particularly ones related to connectors) are still
/// protected under this single lock. So long as this object is held, it is guaranteed that the lock
/// is held.
pub struct ModeConfigGuard<'a, T: KmsDriver> {
    dev: &'a Device<T>,
}

impl<'a, T: KmsDriver> ModeConfigGuard<'a, T> {
    /// Assert that the given device is the owner of this mode config guard.
    ///
    /// # Panics
    ///
    /// Panics if `dev` is different from the owning device for this mode config guard.
    pub(crate) fn assert_owner(&self, dev: &Device<T>) {
        assert!(ptr::eq(self.dev, dev))
    }
}

impl<'a, T: KmsDriver> Drop for ModeConfigGuard<'a, T> {
    fn drop(&mut self) {
        // SAFETY: `mutex` is initialized for as long as we expose `Device<T>` objects to users
        unsafe { bindings::mutex_unlock(&mut (*self.dev.as_raw()).mode_config.mutex) }
    }
}

pub(super) struct UnsafeModeConfigGuard<'a, T: KmsDriver> {
    inner: ManuallyDrop<ModeConfigGuard<'a, T>>
}

impl<'a, T: KmsDriver> UnsafeModeConfigGuard<'a, T> {
    pub(super) unsafe fn new(dev: &'a Device<T>) -> Self {
        Self { inner: ManuallyDrop::new(ModeConfigGuard { dev }) }
    }
}

impl<'a, T: KmsDriver> Deref for UnsafeModeConfigGuard<'a, T> {
    type Target = ModeConfigGuard<'a, T>;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}
