// SPDX-License-Identifier: GPL-2.0 OR MIT

//! Atomic modesetting abstractions for rust.
use alloc::boxed::Box;
use crate::{
    bindings,
    drm::{device::Device, drv::Driver},
    types::*,
    error::{from_err_ptr, to_result, from_result},
    init::Zeroable,
    prelude::*,
    private::Sealed
};
use core::{
    marker::*,
    ptr::{NonNull, addr_of_mut},
    cell::Cell,
    ffi::*,
    slice,
    ops::{Deref, DerefMut},
    mem::ManuallyDrop,
    iter::Iterator,
};
use super::{
    crtc::*,
    connector::*,
    plane::*,
    KmsDriver,
    ModeObject
};

/// The base type for an atomic modesetting state for KMS.
///
/// Generally, a driver will only interact with this type after the state has been committed to
/// hardware - and will go through the `AtomicStateMutator` or `AtomicStateComposer` types when
/// modifying a state prior to/during commit.
#[repr(transparent)]
pub struct AtomicState<T: KmsDriver> {
    pub(super) state: Opaque<bindings::drm_atomic_state>,
    _p: PhantomData<T>,
}

impl<T: KmsDriver> AtomicState<T> {
    /// Reconstruct an immutable reference to an atomic state from the given pointer
    ///
    /// SAFETY: `ptr` must point to a valid instance of an atomic state
    pub(super) unsafe fn from_raw<'a>(ptr: *const bindings::drm_atomic_state) -> &'a Self {
        // SAFETY: Our data layout is identical
        unsafe { &*ptr.cast() }
    }

    pub fn drm_dev(&self) -> &Device<T> {
        // SAFETY: `AtomicState` is initialized for as long as its exposed to users, and `dev` is
        // static throughout the lifetime of `AtomicState`
        unsafe { Device::borrow((*self.state.get()).dev) }
    }

    pub fn get_old_crtc_state<C, S>(&self, crtc: &C) -> Option<&S>
    where
        C: AsRawCrtc<State = S> + ModeObject<Driver = T>,
        S: FromRawCrtcState<Crtc = C>
    {
        // SAFETY: We're guaranteed this is of type `C::State` due to type invariance
        unsafe {
            bindings::drm_atomic_get_old_crtc_state(
                self.state.get(),
                crtc.as_raw()
            ).as_ref().map(|p| unsafe { S::from_raw(p) })
        }
    }

    pub fn get_old_plane_state<P, S>(&self, plane: &P) -> Option<&S>
    where
        P: AsRawPlane<State = S> + ModeObject<Driver = T>,
        S: FromRawPlaneState<Driver = T, Plane = P>,
    {
        // SAFETY: We're guaranteed this is of type `P::State` due to type invariance
        unsafe {
            bindings::drm_atomic_get_old_plane_state(
                self.state.get(),
                plane.as_raw()
            ).as_ref().map(|p| unsafe { S::from_raw(p) })
        }
    }

    pub fn get_old_connector_state<C, S>(&self, connector: &C) -> Option<&S>
    where
        C: AsRawConnector<State = S> + ModeObject<Driver = T>,
        S: FromRawConnectorState<Driver = T, Connector = C>
    {
        // SAFETY: We're guaranteed this is of type `C::State` due to type invariance
        unsafe {
            bindings::drm_atomic_get_old_connector_state(
                self.state.get(),
                connector.as_raw()
            ).as_ref().map(|p| unsafe { S::from_raw(p) })
        }
    }

    pub fn get_new_crtc_state<C, S>(&self, crtc: &C) -> Option<&S>
    where
        C: AsRawCrtc<State = S> + ModeObject<Driver = T>,
        S: FromRawCrtcState<Crtc = C>
    {
        // SAFETY: We're guaranteed this is of type `C::State` due to type invariance
        unsafe {
            bindings::drm_atomic_get_new_crtc_state(
                self.state.get(),
                crtc.as_raw()
            ).as_ref().map(|p| unsafe { S::from_raw(p) })
        }
    }

    pub fn get_new_plane_state<P, S>(&self, plane: &P) -> Option<&S>
    where
        P: AsRawPlane<State = S> + ModeObject<Driver = T>,
        S: FromRawPlaneState<Plane = P, Driver = T>
    {
        // SAFETY: We're guaranteed this is of type `P::State` due to type invariance
        unsafe {
            bindings::drm_atomic_get_new_plane_state(
                self.state.get(),
                plane.as_raw(),
            ).as_ref().map(|p| unsafe { S::from_raw(p) })
        }
    }

    pub fn get_new_connector_state<C, S>(&self, connector: &C) -> Option<&S>
    where
        C: AsRawConnector<State = S> + ModeObject<Driver = T>,
        S: FromRawConnectorState<Connector = C, Driver = T>,
    {
        // SAFETY: We're guaranteed this is of type `C::State` due to type invariance
        unsafe {
            bindings::drm_atomic_get_new_connector_state(
                self.state.get(),
                connector.as_raw()
            ).as_ref().map(|p| unsafe { S::from_raw(p) })
        }
    }
}

// SAFETY: DRM atomic state objects are always reference counted and the get/put functions satisfy
// the requirements.
unsafe impl<T: KmsDriver> AlwaysRefCounted for AtomicState<T> {
    fn inc_ref(&self) {
        // SAFETY: FFI call with no special requirements
        unsafe { bindings::drm_atomic_state_get(self.state.get()) }
    }

    unsafe fn dec_ref(obj: NonNull<Self>) {
        // SAFETY: FFI calls with no special requirements
        unsafe { bindings::drm_atomic_state_put(obj.as_ptr().cast()) }
    }
}

/// A smart-pointer for modifying the contents of an atomic state.
///
/// As it's not unreasonable for a modesetting driver to want to have references to the state of
/// multiple modesetting objects at once, along with mutating multiple states for unique modesetting
/// objects at once, this type provides a mechanism for safely doing both of these things.
///
/// To honor Rust's aliasing rules regarding mutable references, this structure ensures only one
/// mutable reference to a mode object's atomic state may exist at a time - and refuses to provide
/// another if one has already been taken out using runtime checks.
pub struct AtomicStateMutator<T: KmsDriver> {
    /// The state being mutated. Note that the use of `ManuallyDrop` here is because mutators are
    /// only constructed in FFI callbacks and thus borrow their references to the atomic state from
    /// DRM. Composers, which make use of mutators internally, can potentially be owned by rust code
    /// if a driver is performing an atomic commit internally - and thus will call the drop
    /// implementation here.
    state: ManuallyDrop<ARef<AtomicState<T>>>,

    /// Bitmask of borrowed CRTC state objects
    pub(super) borrowed_crtcs: Cell<u32>,
    /// Bitmask of borrowed plane state objects
    pub(super) borrowed_planes: Cell<u32>,
    /// Bitmask of borrowed connector state objects
    pub(super) borrowed_connectors: Cell<u32>,
}

impl<T: KmsDriver> AtomicStateMutator<T> {
    pub(super) unsafe fn from_raw(ptr: NonNull<bindings::drm_atomic_state>) -> Self {
        Self {
            // SAFETY: The data layout of `AtomicState<T>` is identical to drm_atomic_state
            state: ManuallyDrop::new(unsafe { ARef::from_raw(ptr.cast()) }),
            borrowed_planes: Cell::default(),
            borrowed_crtcs: Cell::default(),
            borrowed_connectors: Cell::default(),
        }
    }

    /// Retrieve the last committed atomic state for `crtc` if `crtc` has already been added to the
    /// atomic state being composed.
    ///
    /// Returns `None` otherwise.
    pub fn get_old_crtc_state<C, S>(&self, crtc: &C) -> Option<&S>
    where
        C: AsRawCrtc<State = S> + ModeObject<Driver = T>,
        S: FromRawCrtcState<Crtc = C>
    {
        self.state.get_old_crtc_state(crtc)
    }

    /// Retrieve the last committed atomic state for `connector` if `connector` has already been
    /// added to the atomic state being composed.
    ///
    /// Returns `None` otherwise.
    pub fn get_old_connector_state<C, S>(&self, connector: &C) -> Option<&S>
    where
        C: AsRawConnector<State = S> + ModeObject<Driver = T>,
        S: FromRawConnectorState<Driver = T, Connector = C>
    {
        self.state.get_old_connector_state(connector)
    }

    /// Retrieve the last committed atomic state for `plane` if `plane` has already been added to
    /// the atomic state being composed.
    ///
    /// Returns `None` otherwise.
    pub fn get_old_plane_state<P, S>(&self, plane: &P) -> Option<&S>
    where
        P: AsRawPlane<State = S> + ModeObject<Driver = T>,
        S: FromRawPlaneState<Plane = P, Driver = T>
    {
        self.state.get_old_plane_state(plane)
    }

    /// Return a composer for `plane`s new atomic state if it was previously added to the atomic
    /// state being composed.
    ///
    /// Returns `None` otherwise, or if a composer still exists for this state.
    pub fn get_new_crtc_state<C, S>(&self, crtc: &C) -> Option<BorrowedCrtcState<'_, S>>
    where
        C: AsRawCrtc<State = S>,
        S: FromRawCrtcState<Crtc = C>
    {
        // SAFETY: FFI call with no special requirements
        // Note the only possible return values from DRM here are a valid pointer to the crtc state,
        // or NULL
        let state = unsafe {
            bindings::drm_atomic_get_new_crtc_state(
                self.state.state.get(),
                crtc.as_raw()
            )
        };

        BorrowedCrtcState::new(self, NonNull::new(state)?)
    }

    /// Return a composer for `plane`s new atomic state if it was previously added to the atomic
    /// state being composed.
    ///
    /// Returns `None` otherwise, or if a composer still exists for this state.
    pub fn get_new_plane_state<P, S>(&self, plane: &P) -> Option<BorrowedPlaneState<'_, S>>
    where
        P: AsRawPlane<Driver = T, State = S>,
        S: FromRawPlaneState<Plane = P, Driver = T>
    {
        // SAFETY: FFI call with no special requirements
        // Note the only possible return values from DRM here are a valid pointer to the plane
        // state, or NULL
        let state = unsafe {
            bindings::drm_atomic_get_new_plane_state(
                self.state.state.get(),
                plane.as_raw()
            )
        };

        BorrowedPlaneState::<S>::new(self, NonNull::new(state)?)
    }

    /// Return a composer for `crtc`s new atomic state if it was previously added to the atomic
    /// state being composed.
    ///
    /// Returns `None` otherwise, or if a composer still exists for this state.
    pub fn get_new_connector_state<C, S>(
        &self,
        connector: &C
    ) -> Option<BorrowedConnectorState<'_, S>>
    where
        C: AsRawConnector<State = S> + ModeObject<Driver = T>,
        S: FromRawConnectorState<Driver = T, Connector = C>
    {
        // SAFETY: FFI call with no special requirements
        // Note the only possible return values from DRM here are a valid pointer to the connector
        // state, or NULL
        let state = unsafe {
            bindings::drm_atomic_get_new_connector_state(
                self.state.state.get(),
                connector.as_raw()
            )
        };

        BorrowedConnectorState::new(self, NonNull::new(state)?)
    }

    /// Iterate through each of the planes (regardless of type) currently in this atomic state.
    pub fn iter_planes(&self) -> AtomicStatePlaneIter<'_, T> {
        AtomicStatePlaneIter::new(&self.state)
    }
}

/// An owned atomic state which is not yet part of any commit operation.
///
/// Since it's not yet part of a commit operation, new mode objects may be added to the state. This
/// type is guaranteed to be the only reference to the atomic state for the duration of its lifetime
/// This is usually until a commit operation for this state begins, or this object is dropped.
pub struct AtomicStateComposer<T: KmsDriver> {
    inner: AtomicStateMutator<T>
}

impl<T: KmsDriver> Deref for AtomicStateComposer<T> {
    type Target = AtomicStateMutator<T>;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl<T: KmsDriver> Drop for AtomicStateComposer<T> {
    fn drop(&mut self) {
        // SAFETY: We're in drop, so this is guaranteed to be the last possible reference
        unsafe { ManuallyDrop::drop(&mut self.inner.state) }
    }
}

impl<T: KmsDriver> AtomicStateComposer<T> {
    unsafe fn from_raw(ptr: NonNull<bindings::drm_atomic_state>) -> Self {
        // SAFETY: see `AtomicStateMutator::from_raw()`
         Self { inner: unsafe { AtomicStateMutator::from_raw(ptr) } }
    }

    /// Attempt to add the state for `crtc` to the atomic state for this composer if it hasn't
    /// already been added, and create a mutator for it.
    ///
    /// If a composer already exists for this `crtc`, this function returns `Error(EBUSY)`. If
    /// attempting to add the state fails, another error code will be returned.
    pub fn add_crtc_state<C, S>(&self, crtc: &C) -> Result<BorrowedCrtcState<'_, S>>
    where
        C: AsRawCrtc<State = S> + ModeObject<Driver = T>,
        S: FromRawCrtcState<Crtc = C>
    {
        // SAFETY: We're guaranteed this is of type `C::State` due to type invariance, and we're
        // guaranteed by DRM that this function can only either return a valid pointer to a CRTC
        // state - or an error.
        let state = unsafe {
            from_err_ptr(
                bindings::drm_atomic_get_crtc_state(
                    self.state.state.get(),
                    crtc.as_raw()
                )
            ).map(|c| NonNull::new_unchecked(c))
        }?;

        BorrowedCrtcState::new(self, state).ok_or(EBUSY)
    }

    /// Attempt to add the state for `plane` to the atomic state for this composer if it hasn't
    /// already been added, and create a mutator for it.
    ///
    /// If a composer already exists for this `plane`, this function returns `Error(EBUSY)`. If
    /// attempting to add the state fails, another error code will be returned.
    pub fn add_plane_state<P, S>(&self, plane: &P) -> Result<BorrowedPlaneState<'_, S>>
    where
        P: AsRawPlane<State = S> + ModeObject<Driver = T>,
        S: FromRawPlaneState<Driver = T, Plane = P>,
    {
        // SAFETY: We're guaranteed this is of type `C::State` due to type invariance, and our
        // interface guarantees that this is the only possible reference to the plane's state
        let state = unsafe {
            from_err_ptr(
                bindings::drm_atomic_get_plane_state(
                    self.state.state.get(),
                    plane.as_raw()
                )
            ).map(|p| NonNull::new_unchecked(p))
        }?;

        BorrowedPlaneState::<S>::new(self, state).ok_or(EBUSY)
    }

    /// Attempt to add the state for `connector` to the atomic state for this composer if it hasn't
    /// already been added, and create a mutator for it.
    ///
    /// If a composer already exists for this `connector`, this function returns `Error(EBUSY)`. If
    /// attempting to add the state fails, another error code will be returned.
    pub fn add_connector_state<C, S>(
        &self,
        connector: &C
    ) -> Result<BorrowedConnectorState<'_, S>>
    where
        C: AsRawConnector<State = S> + ModeObject<Driver = T>,
        S: FromRawConnectorState<Driver = T, Connector = C>
    {
        // SAFETY: We're guaranteed this is of type `C::State` due to type invariance, and we're
        // guaranteed by DRM that this function can only either return a valid pointer to a
        // connector state - or an error.
        let state = unsafe {
            from_err_ptr(
                bindings::drm_atomic_get_connector_state(
                    self.state.state.get(),
                    connector.as_raw(),
                )
            ).map(|c| NonNull::new_unchecked(c))
        }?;

        BorrowedConnectorState::new(self, state).ok_or(EBUSY)
    }

    /// Attempt to add any planes affected by changes on the provided CRTC to this atomic state. This
    /// returns an error code if doing so fails.
    pub fn add_affected_planes(&self, crtc: &impl AsRawCrtc) -> Result {
        // SAFETY: FFI call with no special safety requirements
        to_result(unsafe {
            bindings::drm_atomic_add_affected_planes(self.state.state.get(), crtc.as_raw())
        })
    }
}

/// An atomic state composer primarily intended for usage in FFI callbacks. It does not drop a
/// refcount for the atomic state upon being dropped.
pub(super) struct LeakyAtomicStateComposer<T: KmsDriver> {
    composer: ManuallyDrop<AtomicStateComposer<T>>
}

impl<T: KmsDriver> LeakyAtomicStateComposer<T> {
    pub(super) unsafe fn new(state: NonNull<bindings::drm_atomic_state>) -> Self {
        Self {
            // SAFETY: Guaranteed safe by type invariance
            composer: ManuallyDrop::new(unsafe { AtomicStateComposer::from_raw(state) })
        }
    }
}

impl<T: KmsDriver> Deref for LeakyAtomicStateComposer<T> {
    type Target = AtomicStateComposer<T>;

    fn deref(&self) -> &Self::Target {
        &self.composer
    }
}

/// An iterator which goes through each DRM plane currently in an atomic state.
///
/// Note that this iterator will return `OpaquePlane`s, because it's entirely possible for a driver
/// to have multiple implementations of `DriverPlane` - so we don't know what the fully qualified
/// type of each plane is.
pub struct AtomicStatePlaneIter<'a, T: KmsDriver> {
    state: &'a AtomicState<T>,
    current_idx: u8,
}

impl<'a, T: KmsDriver> Iterator for AtomicStatePlaneIter<'a, T> {
    type Item = &'a OpaquePlane<T>;

    fn next(&mut self) -> Option<Self::Item> {
        let ptr = self.state.state.get();

        // SAFETY: `planes` is initialized by the time we expose AtomicState<T> through any form to
        // users. And because we don't allow state mutation outside of mutators, which are single
        // threaded, the contents of this struct are at least guaranteed not to change through the
        // duration of this borrow.
        let planes: &[bindings::__drm_planes_state] = unsafe {
            slice::from_raw_parts((*ptr).planes.cast_const(), self.state.drm_dev().num_plane() as _)
        };

        for plane_states in &planes[self.current_idx as _..] {
            self.current_idx += 1;
            if !plane_states.ptr.is_null() {
                // SAFETY: OpaquePlane has an identical data layout, and its only possible values
                // are NULL or pointing at a valid drm_plane
                return Some(unsafe { OpaquePlane::from_raw(plane_states.ptr) });
            }
        }

        None
    }
}

impl<'a, T: KmsDriver> AtomicStatePlaneIter<'a, T> {
    fn new(state: &'a AtomicState<T>) -> AtomicStatePlaneIter<'a, T> {
        Self {
            current_idx: 0,
            state
        }
    }
}
