// SPDX-License-Identifier: GPL-2.0 OR MIT

//! Rust bindings for DRM connectors

use crate::{
    bindings,
    sync::ArcBorrow,
    drm::{
        drv::{Driver, FEAT_MODESET},
        device::Device,
    },
    types::{AlwaysRefCounted, Opaque, ARef},
    prelude::*,
    init::Zeroable,
    error::{to_result, from_result},
    build_error,
    private::Sealed,
};
use core::{
    marker::*,
    ptr::null_mut,
    mem,
    ptr::{self, NonNull, addr_of_mut},
    ffi::*,
    ops::*,
    cell::Cell,
};
use super::{
    ModeObject,
    RcModeObject,
    ModeConfigGuard,
    UnsafeModeConfigGuard,
    encoder::*,
    KmsDriver,
    atomic::*,
};
use macros::pin_data;

// XXX: This is :\, figure out a better way at some point?
pub use bindings::{
    DRM_MODE_CONNECTOR_Unknown,
    DRM_MODE_CONNECTOR_VGA,
    DRM_MODE_CONNECTOR_DVII,
    DRM_MODE_CONNECTOR_DVID,
    DRM_MODE_CONNECTOR_DVIA,
    DRM_MODE_CONNECTOR_Composite,
    DRM_MODE_CONNECTOR_SVIDEO,
    DRM_MODE_CONNECTOR_LVDS,
    DRM_MODE_CONNECTOR_Component,
    DRM_MODE_CONNECTOR_9PinDIN,
    DRM_MODE_CONNECTOR_DisplayPort,
    DRM_MODE_CONNECTOR_HDMIA,
    DRM_MODE_CONNECTOR_HDMIB,
    DRM_MODE_CONNECTOR_TV,
    DRM_MODE_CONNECTOR_eDP,
    DRM_MODE_CONNECTOR_VIRTUAL,
    DRM_MODE_CONNECTOR_DSI,
    DRM_MODE_CONNECTOR_DPI,
    DRM_MODE_CONNECTOR_WRITEBACK,
    DRM_MODE_CONNECTOR_SPI,
    DRM_MODE_CONNECTOR_USB,
};

/// A DRM connector implementation
pub trait DriverConnector: Send + Sync + Sized {
    /// The data type to use for passing incoming arguments for new `Connector<T>` instances
    /// Drivers which don't care about this can just use `()`
    type Args;

    /// The parent driver for this DRM connector implementation
    type Driver: KmsDriver;

    /// The atomic state implementation for this DRM connector implementation
    type State: DriverConnectorState;

    /// Create a new instance of the private driver data struct for this connector in-place
    fn new(device: &Device<Self::Driver>, args: Self::Args) -> impl PinInit<Self, Error>;

    /// Retrieve a list of available display modes for this connector
    fn get_modes<'a>(
        connector: ConnectorGuard<'a, Self>,
        guard: &ModeConfigGuard<'a, Self::Driver>
    ) -> i32;
}

/// A typed DRM connector
#[repr(C)]
#[pin_data]
pub struct Connector<T: DriverConnector> {
    connector: Opaque<bindings::drm_connector>,
    #[pin]
    inner: T,
    #[pin]
    _p: PhantomPinned
}

impl<T: DriverConnector> Sealed for Connector<T> { }

// SAFETY: DRM expects this struct to be zero-initialized
unsafe impl Zeroable for bindings::drm_connector { }

impl<T: DriverConnector> Deref for Connector<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl<T: DriverConnector> Connector<T> {
    const FUNCS: bindings::drm_connector_funcs = bindings::drm_connector_funcs {
        dpms: None,
        atomic_get_property: None,
        atomic_set_property: None,
        early_unregister: None,
        late_register: None,
        set_property: None,
        reset: Some(connector_reset_callback::<T::State>),
        atomic_print_state: None,
        atomic_destroy_state: Some(atomic_destroy_state_callback::<T::State>),
        destroy: Some(connector_destroy_callback::<T>),
        force: None,
        detect: None,
        fill_modes: Some(bindings::drm_helper_probe_single_connector_modes),
        debugfs_init: None,
        oob_hotplug_event: None,
        atomic_duplicate_state: Some(atomic_duplicate_state_callback::<T::State>),
    };

    const HELPER_FUNCS: bindings::drm_connector_helper_funcs = bindings::drm_connector_helper_funcs {
        mode_valid: None,
        atomic_check: None,
        get_modes: Some(get_modes_callback::<T>),
        detect_ctx: None,
        enable_hpd: None,
        disable_hpd: None,
        best_encoder: None,
        atomic_commit: None,
        mode_valid_ctx: None,
        atomic_best_encoder: None,
        prepare_writeback_job: None,
        cleanup_writeback_job: None,
    };

    pub fn new(
        dev: &Device<T::Driver>,
        type_: u32,
        args: T::Args,
    ) -> Result<ARef<Self>> {
        let new: Pin<Box<Self>> = Box::try_pin_init(
            try_pin_init!(Self {
                connector: Opaque::new(bindings::drm_connector {
                    helper_private: &Self::HELPER_FUNCS,
                    ..Default::default()
                }),
                inner <- T::new(dev, args),
                _p: PhantomPinned
            }),
            GFP_KERNEL
        )?;

        // SAFETY: FFI call with no special safety requirements
        to_result(unsafe {
            bindings::drm_connector_init(
                dev.as_raw(),
                new.as_raw(),
                &Self::FUNCS,
                type_ as i32
            )
        })?;

        // Convert the connector into an ARef so the caller has proper ownership over a refcount to
        // it. Also, the Box we consume here will be reconstructed in connector_destroy_callback()
        // once the connector's refcount drops to zero.
        // SAFETY: We currently hold ownership of the Box containing the connector and it's
        // refcount. As well, this operation will not move the contents of the Box.
        Ok(unsafe {
            ARef::from_raw(NonNull::new_unchecked(Box::into_raw(Pin::into_inner_unchecked(new))))
        })
    }

    /// Acquire a `ConnectorGuard` for this connector from a `ModeConfigGuard`.
    ///
    /// This verifies using the provided reference that the given guard is actually for the same
    /// device as this connector's parent.
    ///
    /// # Panics
    ///
    /// Panics if `guard` is not a mode config guard for this connector's parent DRM device
    pub fn guard<'a>(&'a self, guard: &ModeConfigGuard<'a, T::Driver>) -> ConnectorGuard<'a, T> {
        guard.assert_owner(self.drm_dev());
        ConnectorGuard { connector: self }
    }

    /// Attach an encoder to this connector, this should only be done before registration
    #[must_use]
    pub fn attach_encoder<E>(&self, encoder: &Encoder<E>) -> Result
    where
        E: DriverEncoder<Driver = T::Driver>
    {
        // SAFETY: FFI call with no special requirements
        to_result(unsafe {
            bindings::drm_connector_attach_encoder(self.as_raw(), encoder.as_raw())
        })
    }
}

/// An object which can be used as a DRM connector
pub trait AsRawConnector: RcModeObject {
    /// The driver for this DRM connector interface
    type Driver: KmsDriver;

    /// The type this connector interface returns for its atomic state
    type State: AsRawConnectorState;

    /// Return the raw `bindings::drm_connector` for this DRM connector.
    ///
    /// Drivers should never use this directly
    fn as_raw(&self) -> *mut bindings::drm_connector;

    /// Convert a raw `bindings::drm_connector` pointer into an object of this type.
    ///
    /// SAFETY: Callers promise that `ptr` points to a valid instance of this type.
    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_connector) -> &'a Self;
}

// SAFETY: Our connector interfaces are guaranteed to be thread-safe
unsafe impl<T: DriverConnector> Send for Connector<T> {}
unsafe impl<T: DriverConnector> Sync for Connector<T> {}

impl<T: DriverConnector> ModeObject for Connector<T> {
    type Driver = T::Driver;

    fn drm_dev(&self) -> &Device<Self::Driver> {
        // SAFETY: The parent device for a DRM connector will never outlive the connector, and this
        // pointer is invariant through the lifetime of the connector
        unsafe { Device::borrow((*self.as_raw()).dev) }
    }

    fn raw_mode_obj(&self) -> *mut bindings::drm_mode_object {
        // SAFETY: We don't expose DRM connectors to users before `base` is initialized
        unsafe { addr_of_mut!((*self.as_raw()).base) }
    }
}

// SAFETY: DRM connectors are refcounted mode objects
unsafe impl<T: DriverConnector> RcModeObject for Connector<T> {}

impl<T: DriverConnector> AsRawConnector for Connector<T> {
    type Driver = T::Driver;
    type State = ConnectorState<T::State>;

    fn as_raw(&self) -> *mut bindings::drm_connector {
        self.connector.get()
    }

    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_connector) -> &'a Self {
        // SAFETY: Our data layout starts with `bindings::drm_connector`
        unsafe { &*ptr.cast() }
    }
}

/// Methods available on any object which can be used as a DRM connector
///
/// They cannot be overriden.
pub trait RawConnector: AsRawConnector {
    /// Return the index of this DRM connector
    #[inline]
    fn index(&self) -> u32 {
        // SAFETY: The index is initialized by the time we expose DRM connector objects to users,
        // and is invariant throughout the lifetime of the connector
        unsafe { (*self.as_raw()).index }
    }

    /// Return the bitmask derived from this DRM connector's index
    #[inline]
    fn mask(&self) -> u32 {
        1 << self.index()
    }
}
impl<T: AsRawConnector> RawConnector for T {}

unsafe extern "C" fn connector_destroy_callback<T: DriverConnector>(
    connector: *mut bindings::drm_connector,
) {
    // SAFETY: Connector has to point to a valid drm_connector, as its function table is the only
    // existing entrypoint to this function
    unsafe {
        bindings::drm_connector_unregister(connector);
        bindings::drm_connector_cleanup(connector);
    };

    // SAFETY: We always create connectors in boxes, and since we are running from this connector's
    // destructor we are guaranteed to have the last remaining reference. Furthermore, we're
    // guaranteed by type invariance that the contents of the box are of type Connector<T>.
    drop(unsafe { Box::from_raw(connector as *mut Connector<T>) });
}

unsafe extern "C" fn get_modes_callback<T: DriverConnector>(
    connector: *mut bindings::drm_connector,
) -> c_int {
    // SAFETY: We're guaranteed by type invariants that connector is of type Connector<T>, and
    // connector must point to a valid instance of Connector<T> as it's the only entry-point to this
    // callback.
    let connector = unsafe { Connector::<T>::from_raw(connector) };

    // SAFETY: This FFI callback is only called while the mode config lock is held
    let guard = unsafe { UnsafeModeConfigGuard::new(connector.drm_dev()) };

    T::get_modes(connector.guard(&guard), &guard)
}

/// A DRM connector without a known `DriverConnector` implementation
///
/// Used in situations where we can't automatically derive the `DriverConnector` implementation for
/// a DRM connector
#[repr(transparent)]
pub struct OpaqueConnector<T: KmsDriver> {
    connector: Opaque<bindings::drm_connector>,
    _p: PhantomData<T>
}

impl<T: KmsDriver> Sealed for OpaqueConnector<T> {}

impl<T: KmsDriver> AsRawConnector for OpaqueConnector<T> {
    type Driver = T;
    type State = OpaqueConnectorState<T>;

    fn as_raw(&self) -> *mut bindings::drm_connector {
        self.connector.get()
    }

    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_connector) -> &'a Self {
        // SAFETY: Our data layout is identical to `bindings::drm_connector`
        unsafe { &*ptr.cast() }
    }
}

impl<T: KmsDriver> ModeObject for OpaqueConnector<T> {
    type Driver = T;

    fn drm_dev(&self) -> &Device<Self::Driver> {
        // SAFETY: The parent device for a DRM connector will never outlive the connector, and this
        // pointer is invariant through the lifetime of the connector
        unsafe { Device::borrow((*self.as_raw()).dev) }
    }

    fn raw_mode_obj(&self) -> *mut bindings::drm_mode_object {
        // SAFETY: We don't expose DRM connectors to users before `base` is initialized
        unsafe { addr_of_mut!((*self.as_raw()).base) }
    }
}

// SAFETY: Connectors are reference counted mode objects
unsafe impl<T: KmsDriver> RcModeObject for OpaqueConnector<T> {}

// SAFETY: Our connector interfaces are guaranteed to be thread-safe
unsafe impl<T: KmsDriver> Send for OpaqueConnector<T> {}
unsafe impl<T: KmsDriver> Sync for OpaqueConnector<T> {}

/// A privileged smart-pointer for `Connector<T>` which proves that the owner currently is protected
/// under the `bindings::drm_device.mode_config.mutex` lock and provides access to data and methods
/// protected under said lock.
#[derive(Copy, Clone)]
pub struct ConnectorGuard<'a, T: DriverConnector> {
    connector: &'a Connector<T>,
}

impl<T: DriverConnector> Deref for ConnectorGuard<'_, T> {
    type Target = Connector<T>;

    fn deref(&self) -> &Self::Target {
        self.connector
    }
}

impl<'a, T: DriverConnector> ConnectorGuard<'a, T> {
    pub fn add_modes_noedid(&self, (max_h, max_v): (i32, i32)) -> i32 {
        unsafe { bindings::drm_add_modes_noedid(self.as_raw(), max_h, max_v) }
    }

    pub fn set_preferred_mode(&self, (h_pref, w_pref): (i32, i32)) {
        unsafe { bindings::drm_set_preferred_mode(self.as_raw(), h_pref, w_pref) }
    }
}

// SAFETY: DRM expects this struct to be zero-initialized
unsafe impl Zeroable for bindings::drm_connector_state {}

/// A type which which can produce a reference to a raw DRM connector state structure.
///
/// This can be any type that contains a `bindings::drm_connector_state`, or contains a reference to
/// a `bindings::drm_connector_state`. Or in other words - any rust type that can be used as a DRM
/// connector state.
///
/// Note that unlike `AsRawConnector`, not all objects which implement `AsRawConnectorState` can be
/// obtained from raw `bindings::drm_connector_state` pointers.
pub trait AsRawConnectorState: private::AsRawConnectorState {
    type Driver: KmsDriver;
    type Connector: AsRawConnector;
}

pub(super) mod private {
    use super::*;

    /// Trait for retrieving references to the base connector state contained within any connector
    /// state compatible type
    #[doc(hidden)]
    pub trait AsRawConnectorState {
        /// Return an immutable reference to the raw connector state
        fn as_raw(&self) -> &bindings::drm_connector_state;

        /// Get a mutable reference to the raw `bindings::drm_connector_state` contained within this
        /// type.
        ///
        /// SAFETY: The caller promises this mutable reference will not be used to modify any
        /// contents of `bindings::drm_connector_state` which DRM would consider to be static - like
        /// the backpointer to the DRM connector that owns this state. This also means the mutable
        /// reference should never be exposed outside of this crate.
        unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_connector_state;
    }
}

pub(super) use private::AsRawConnectorState as AsRawConnectorStatePrivate;

/// A type which can be obtained directly from a raw DRM connector state pointer.
///
/// This can be any type which can be created directly from a raw pointer to a
/// `bindings::drm_connector_state`, and which can be used as a connector state.
pub trait FromRawConnectorState: AsRawConnectorState {
    /// Get an immutable reference to this type from the given raw `bindings::drm_connector_state`
    /// pointer
    ///
    /// SAFETY: The caller guarantees `ptr` is contained within a valid instance of `Self`
    unsafe fn from_raw<'a>(ptr: *const bindings::drm_connector_state) -> &'a Self;

    /// Get a mutable reference to this type from the given raw `bindings::drm_connector_state`
    /// pointer
    ///
    /// SAFETY: The caller guarantees `ptr` is contained within a valid instance of `Self`, and that
    /// no other references (mutable or immutable) to `ptr` exist.
    unsafe fn from_raw_mut<'a>(ptr: *mut bindings::drm_connector_state) -> &'a mut Self;
}

/// Methods available on any type which can be used as a raw DRM connector state.
pub trait RawConnectorState: AsRawConnectorState {
    fn connector(&self) -> &Self::Connector {
        // SAFETY: This is guaranteed safe by type invariance, and we're guaranteed by DRM that
        // `self.state.connector` points to a valid instance of a `Connector<T>`
        unsafe { Self::Connector::from_raw((*self.as_raw()).connector) }
    }
}
impl<T: AsRawConnectorState> RawConnectorState for T {}

#[derive(Default)]
#[repr(C)]
pub struct ConnectorState<T: DriverConnectorState> {
    state: bindings::drm_connector_state,
    inner: T,
}

/// The trait for a driver's atomic DRM connector state
pub trait DriverConnectorState: Clone + Default + Sized {
    type Connector: DriverConnector;
}

impl<T: DriverConnectorState> Sealed for ConnectorState<T> {}

impl<T: DriverConnectorState> AsRawConnectorState for ConnectorState<T> {
    type Driver = <T::Connector as DriverConnector>::Driver;
    type Connector = Connector<T::Connector>;
}

impl<T: DriverConnectorState> private::AsRawConnectorState for ConnectorState<T> {
    fn as_raw(&self) -> &bindings::drm_connector_state {
        &self.state
    }

    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_connector_state {
        &mut self.state
    }
}

impl<T: DriverConnectorState> FromRawConnectorState for ConnectorState<T> {
    unsafe fn from_raw<'a>(ptr: *const bindings::drm_connector_state) -> &'a Self {
        // SAFETY: Our data layout starts with `bindings::drm_connector_state`
        unsafe { &*ptr.cast() }
    }

    unsafe fn from_raw_mut<'a>(ptr: *mut bindings::drm_connector_state) -> &'a mut Self {
        // SAFETY: Our data layout starts with `bindings::drm_connector_state`
        unsafe { &mut *ptr.cast() }
    }
}

#[repr(transparent)]
pub struct OpaqueConnectorState<T: KmsDriver> {
    state: bindings::drm_connector_state,
    _p: PhantomData<T>
}

impl<T: KmsDriver> AsRawConnectorState for OpaqueConnectorState<T> {
    type Driver = T;
    type Connector = OpaqueConnector<T>;
}

impl<T: KmsDriver> private::AsRawConnectorState for OpaqueConnectorState<T> {
    fn as_raw(&self) -> &bindings::drm_connector_state {
        &self.state
    }

    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_connector_state {
        &mut self.state
    }
}

impl<T: KmsDriver> FromRawConnectorState for OpaqueConnectorState<T> {
    unsafe fn from_raw<'a>(ptr: *const bindings::drm_connector_state) -> &'a Self {
        // SAFETY: Our data layout is identical to `bindings::drm_connector_state`
        unsafe { &*ptr.cast() }
    }

    unsafe fn from_raw_mut<'a>(ptr: *mut bindings::drm_connector_state) -> &'a mut Self {
        // SAFETY: Our data layout is identical to `bindings::drm_connector_state`
        unsafe { &mut *ptr.cast() }
    }
}

/// A smart-pointer for mutating a DRM connector's state
///
/// Returned by an atomic state mutator (see `AtomicStateMutator` and `AtomicStateComposer`) in
/// order to allow for the mutation of a connector state. In order to uphold rust's data-aliasing
/// rules surrounding mutable references, only one `ConnectorStateMutator` may exist for a given
/// object's state at a time.
///
/// As well, drivers may get a direct mutable reference to their private state data through this
/// type's `Deref` implementation.
pub struct BorrowedConnectorState<'a, T: FromRawConnectorState> {
    state: &'a mut T,
    mask: &'a Cell<u32>
}

impl<'a, T: FromRawConnectorState> BorrowedConnectorState<'a, T> {
    pub(super) fn new(
        mutator: &'a AtomicStateMutator<T::Driver>,
        state: NonNull<bindings::drm_connector_state>
    ) -> Option<Self> {
        // SAFETY: `connector` is invariant throughout the lifetime of the atomic state, is
        // initialized by this point, and we're guaranteed it is of type `Connector<T>` by type
        // invariance
        let connector =
            unsafe { OpaqueConnector::<T::Driver>::from_raw((*state.as_ptr()).connector) };
        let conn_mask = connector.mask();
        let borrowed_mask = mutator.borrowed_connectors.get();

        if borrowed_mask & conn_mask == 0 {
            mutator.borrowed_connectors.set(borrowed_mask | conn_mask);
            Some(Self {
                mask: &mutator.borrowed_connectors,
                // SAFETY: We're guaranteed `state` is of `ConnectorState<T>` by type invariance,
                // and we just confirmed by checking `borrowed_connectors` that no other mutable
                // borrows have been taken out for `state`
                state: unsafe { T::from_raw_mut(state.as_ptr()) },
            })
        } else {
            // TODO: Print a kernel warning here, this is a user error
            None
        }
    }
}

impl<'a, T: DriverConnectorState> Deref for BorrowedConnectorState<'a, ConnectorState<T>> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.state.inner
    }
}

impl<'a, T: DriverConnectorState> DerefMut for BorrowedConnectorState<'a, ConnectorState<T>> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.state.inner
    }
}

impl<'a, T: FromRawConnectorState> Drop for BorrowedConnectorState<'a, T> {
    fn drop(&mut self) {
        let mask = self.state.connector().mask();
        self.mask.set(self.mask.get() & !mask);
    }
}

impl<'a, T: FromRawConnectorState> AsRawConnectorState for BorrowedConnectorState<'a, T> {
    type Driver = T::Driver;
    type Connector = T::Connector;
}

impl<'a, T: FromRawConnectorState> private::AsRawConnectorState for BorrowedConnectorState<'a, T> {
    fn as_raw(&self) -> &bindings::drm_connector_state {
        self.state.as_raw()
    }

    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_connector_state {
        // SAFETY: We're bound by the same safety contract as this function
        unsafe { self.state.as_raw_mut() }
    }
}

unsafe extern "C" fn atomic_duplicate_state_callback<T: DriverConnectorState>(
    connector: *mut bindings::drm_connector
) -> *mut bindings::drm_connector_state
{
    // SAFETY: `connector` has to point to a valid instance of drm_connector, since it holds the vtable for
    // this function - which is the only possible entrypoint the caller could have used
    let state = unsafe { (*connector).state };
    if state.is_null() {
        return null_mut();
    }

    // SAFETY: We just verified that `state` is non-null, and we're guaranteed by our bindings that
    // `state` is of type `ConnectorState<T>`.
    let state = unsafe { ConnectorState::<T>::from_raw(state) };

    let new = Box::try_init(
        try_init!(ConnectorState::<T> {
            state: bindings::drm_connector_state { ..Default::default() },
            inner: state.inner.clone()
        }),
        GFP_KERNEL
    );

    if let Ok(mut new) = new {
        // SAFETY: Just a lil' FFI call, nothing special here
        unsafe {
            bindings::__drm_atomic_helper_connector_duplicate_state(connector, new.as_raw_mut())
        };

        Box::into_raw(new).cast()
    } else {
        null_mut()
    }
}

unsafe extern "C" fn atomic_destroy_state_callback<T: DriverConnectorState>(
    _connector: *mut bindings::drm_connector,
    connector_state: *mut bindings::drm_connector_state
) {
    // SAFETY: This callback wouldn't be called unless there a connector state to destroy
    unsafe { bindings::__drm_atomic_helper_connector_destroy_state(connector_state) };

    // SAFETY: We're guaranteed by type invariants that connector_state is of type
    // ConnectorState<T>, and since this is the destructor callback for DRM - we're guaranteed to
    // hold the only remaining reference to this state
    drop(unsafe { Box::from_raw(connector_state.cast::<ConnectorState<T>>()) });
}

unsafe extern "C" fn connector_reset_callback<T: DriverConnectorState>(
    connector: *mut bindings::drm_connector,
) {
    // SAFETY: The only entrypoint to this function lives in `connector` so it must be valid, and
    let state = unsafe { (*connector).state };
    if !state.is_null() {
        // SAFETY: We're guaranteed by type invariance that this connector's state is of type
        // DriverConnectorState<T>
        unsafe { atomic_destroy_state_callback::<T>(connector, state) }

        // SAFETY: No special requirements here, DRM expects this to be NULL
        unsafe { (*connector).state = null_mut(); }
    }

    // Unfortunately, this is the best we can do at the moment as this FFI callback was mistakenly
    // presumed to be infallible :(
    let new = Box::new(ConnectorState::<T>::default(), GFP_KERNEL).expect("Blame the API, sorry!");

    // SAFETY: DRM takes ownership of the state from here and assigns it to the connector
    unsafe {
        bindings::__drm_atomic_helper_connector_reset(connector, Box::into_raw(new).cast())
    };
}
