// SPDX-License-Identifier: GPL-2.0 OR MIT

//! KMS driver abstractions for rust.

use super::{
    atomic::*,
    plane::*,
    ModeObject,
    StaticModeObject,
    KmsDriver
};
use crate::{
    bindings,
    drm::{drv::Driver, device::{Device, PendingDevice}},
    device,
    prelude::*,
    private::Sealed,
    error::from_result,
    types::Opaque,
    init::Zeroable,
    sync::Arc,
    error::to_result,
};
use core::{
    cell::Cell,
    marker::*,
    ptr::{NonNull, null, null_mut, addr_of_mut},
    ops::{Deref, DerefMut},
};
use macros::vtable;

/// KMS CRTC object functions, which must be implemented by drivers.
#[vtable]
pub trait DriverCrtc: Send + Sync + Sized {
    /// The data type to use for passing incoming arguments for new `Crtc<T>` instances
    /// Drivers which don't care about this can just use `()`
    type Args;

    /// The parent driver implementation
    type Driver: KmsDriver;

    /// The type for this driver's drm_crtc_state implementation
    type State: DriverCrtcState;

    /// Create a new CRTC for this driver
    fn new(device: &Device<Self::Driver>, args: Self::Args) -> impl PinInit<Self, Error>;

    fn atomic_check(crtc: &Crtc<Self>, state: &AtomicStateComposer<Self::Driver>) -> Result {
        build_error::build_error("This should not be reachable")
    }
}

/// A typed KMS CRTC with a specific driver.
#[repr(C)]
#[pin_data]
pub struct Crtc<T: DriverCrtc> {
    // The FFI drm_crtc object
    crtc: Opaque<bindings::drm_crtc>,
    /// The driver's private inner data
    #[pin]
    inner: T,
    #[pin]
    _p: PhantomPinned,
}

unsafe impl Zeroable for bindings::drm_crtc { }

impl<T: DriverCrtc> Sealed for Crtc<T> {}

// SAFETY: Our CRTC interfaces are guaranteed to be thread-safe
unsafe impl<T: DriverCrtc> Send for Crtc<T> { }
unsafe impl<T: DriverCrtc> Sync for Crtc<T> { }

impl<T: DriverCrtc> Deref for Crtc<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl<T: DriverCrtc> ModeObject for Crtc<T> {
    type Driver = T::Driver;

    fn drm_dev(&self) -> &Device<Self::Driver> {
        // SAFETY: DRM connectors exist for as long as the device does, so this pointer is always
        // valid
        unsafe { Device::borrow((*self.as_raw()).dev) }
    }

    fn raw_mode_obj(&self) -> *mut bindings::drm_mode_object {
        // SAFETY: We don't expose Crtc<T> to users before it's initialized, so `base` is always
        // initialized
        unsafe { addr_of_mut!((*self.as_raw()).base) }
    }
}

// SAFETY: CRTCs are non-refcounted modesetting objects
unsafe impl<T: DriverCrtc> StaticModeObject for Crtc<T> { }

impl<T: DriverCrtc> Crtc<T> {
    /// The actual C vtable for drm_crtc_funcs
    const FUNCS: bindings::drm_crtc_funcs = bindings::drm_crtc_funcs {
        atomic_destroy_state: Some(atomic_destroy_state_callback::<T::State>),
        atomic_duplicate_state: Some(atomic_duplicate_state_callback::<T::State>),
        atomic_get_property: None,
        atomic_print_state: None,
        atomic_set_property: None,
        cursor_move: None,
        cursor_set2: None,
        cursor_set: None,
        destroy: Some(crtc_destroy_callback::<T>),
        disable_vblank: None,
        early_unregister: None,
        enable_vblank: None,
        gamma_set: None, // TODO
        get_crc_sources: None,
        get_vblank_counter: None,
        get_vblank_timestamp: None,
        late_register: None,
        page_flip: Some(bindings::drm_atomic_helper_page_flip),
        page_flip_target: None,
        reset: Some(crtc_reset_callback::<T::State>),
        set_config: Some(bindings::drm_atomic_helper_set_config),
        set_crc_source: None,
        set_property: None,
        verify_crc_source: None,
    };

    /// The actual C vtable for drm_crtc_helper_funcs
    const HELPER_FUNCS: bindings::drm_crtc_helper_funcs = bindings::drm_crtc_helper_funcs {
        atomic_disable: None,
        atomic_enable: None,
        atomic_check: if T::HAS_ATOMIC_CHECK { Some(atomic_check_callback::<T>) } else { None },
        dpms: None,
        commit: None,
        prepare: None,
        disable: None,
        mode_set: None,
        mode_valid: None,
        mode_fixup: None,
        atomic_begin: None,
        atomic_flush: None,
        mode_set_nofb: None,
        mode_set_base: None,
        mode_set_base_atomic: None,
        get_scanout_position: None,
    };

    pub fn new<'a, 'b: 'a, P, C>(
        dev: &'a PendingDevice<T::Driver>,
        primary: &'a Plane<P>,
        cursor: Option<&'a Plane<C>>,
        name: Option<&CStr>,
        args: T::Args,
    ) -> Result<&'b Self>
    where
        P: DriverPlane,
        C: DriverPlane
    {
        let this = Box::try_pin_init(
            try_pin_init!(Self {
                crtc: Opaque::new(bindings::drm_crtc {
                    helper_private: &Self::HELPER_FUNCS,
                    ..Default::default()
                }),
                inner <- T::new(dev, args),
                _p: PhantomPinned,
            }),
            GFP_KERNEL
        )?;

        to_result(unsafe {
            bindings::drm_crtc_init_with_planes(
                dev.as_raw(),
                this.as_raw(),
                primary.as_raw(),
                cursor.map_or(null_mut(), |c| c.as_raw()),
                &Self::FUNCS,
                name.map_or(null(), |n| n.as_char_ptr())
            )
        })?;

        // Convert the box into a raw pointer, we'll re-assemble it in crtc_destroy_callback()
        // SAFETY: We don't move anything
        Ok(unsafe { &*Box::into_raw(Pin::into_inner_unchecked(this)) })
    }
}

pub trait AsRawCrtc: StaticModeObject {
    /// The type that should be returned for a CRTC state acquired using this CRTC interface
    type State: AsRawCrtcState;

    /// Return a raw pointer to the `bindings::drm_crtc` for this object
    fn as_raw(&self) -> *mut bindings::drm_crtc;

    /// Convert a raw `bindings::drm_crtc` pointer into an object of this type.
    ///
    /// SAFETY: Callers promise that `ptr` points to a valid instance of this type
    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_crtc) -> &'a Self;
}

/// A trait for objects which can be used as a DRM CRTC.
impl<T: DriverCrtc> AsRawCrtc for Crtc<T> {
    type State = CrtcState<T::State>;

    fn as_raw(&self) -> *mut bindings::drm_crtc {
        self.crtc.get()
    }

    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_crtc) -> &'a Self {
        // SAFETY: Our data layout starts with `bindings::drm_crtc`
        unsafe { &*ptr.cast() }
    }
}

/// Common methods which are accessible on any object which can be used as a DRM CRTC.
pub trait RawCrtc: AsRawCrtc {
    /// Return the index of this CRTC
    fn index(&self) -> u32 {
        // SAFETY: The index is initialized by the time we expose Crtc objects to users, and is
        // invariant throughout the lifetime of the Crtc
        unsafe { (*self.as_raw()).index }
    }

    fn mask(&self) -> u32 {
        1 << self.index()
    }
}
impl<T: AsRawCrtc> RawCrtc for T {}

/// A DRM CRTC without a known `CrtcDriver` implementation
///
/// Mainly for situations where our bindings can't infer the `CrtcDriver` implementation
/// automatically. Drivers can use `Crtc::<CrtcDriver>::from_opaque` to attempt upcast the opaque
/// CRTC at runtime.
#[repr(transparent)]
pub struct OpaqueCrtc<T: KmsDriver> {
    crtc: Opaque<bindings::drm_crtc>,
    _p: PhantomData<T>
}

impl<T: KmsDriver> Sealed for OpaqueCrtc<T> {}

impl<T: KmsDriver> AsRawCrtc for OpaqueCrtc<T> {
    type State = OpaqueCrtcState<T>;

    fn as_raw(&self) -> *mut bindings::drm_crtc {
        self.crtc.get()
    }

    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_crtc) -> &'a Self {
        // SAFETY: Our data layout starts with `bindings::drm_crtc`
        unsafe { &*ptr.cast() }
    }
}

impl<T: KmsDriver> ModeObject for OpaqueCrtc<T> {
    type Driver = T;

    fn drm_dev(&self) -> &Device<Self::Driver> {
        // SAFETY: The parent device for a DRM connector will never outlive the connector, and this
        // pointer is invariant through the lifetime of the connector
        unsafe { Device::borrow((*self.as_raw()).dev) }
    }

    fn raw_mode_obj(&self) -> *mut bindings::drm_mode_object {
        // SAFETY: We don't expose DRM connectors to users before `base` is initialized
        unsafe { addr_of_mut!((*self.as_raw()).base) }
    }
}

// SAFETY: CRTCs are non-refcounted modesetting objects
unsafe impl<T: KmsDriver> StaticModeObject for OpaqueCrtc<T> {}

// SAFETY: Our CRTC interface is guaranteed to be thread-safe
unsafe impl<T: KmsDriver> Send for OpaqueCrtc<T> {}
unsafe impl<T: KmsDriver> Sync for OpaqueCrtc<T> {}

unsafe impl Zeroable for bindings::drm_crtc_state { }

impl<T: DriverCrtcState> Sealed for CrtcState<T> {}

pub trait DriverCrtcState: Clone + Default + Sized {
    type Crtc: DriverCrtc;
}

#[derive(Default)]
#[repr(C)]
pub struct CrtcState<T: DriverCrtcState> {
    state: bindings::drm_crtc_state,
    inner: T,
}

impl<T: DriverCrtcState> Deref for CrtcState<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl<T: DriverCrtcState> DerefMut for CrtcState<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.inner
    }
}

/// A trait for objects which can be used as an atomic state for a CRTC
pub trait AsRawCrtcState: private::AsRawCrtcState {
    /// The type that this CRTC state interface returns to represent the parent CRTC
    type Crtc: AsRawCrtc;
}

pub(crate) mod private {
    use super::*;

    #[doc(hidden)]
    pub trait AsRawCrtcState {
        /// Return an immutable reference to the raw CRTC state
        fn as_raw(&self) -> &bindings::drm_crtc_state;

        /// Get a mutable reference to the raw `bindings::drm_crtc_state` contained within this
        /// type.
        ///
        /// SAFETY: The caller promises this mutable reference will not be used to modify any
        /// contents of `bindings::drm_crtc_state` which DRM would consider to be static - like the
        /// backpointer to the DRM plane that owns this state. This also means the mutable reference
        /// should never be exposed outside of this crate.
        unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_crtc_state;
    }
}

pub(super) use private::AsRawCrtcState as AsRawCrtcStatePrivate;

/// A trait for providing common methods which can be used on any type that can be used as an atomic
/// CRTC state.
pub trait RawCrtcState: AsRawCrtcState {
    fn crtc(&self) -> &Self::Crtc {
        // SAFETY: This is guaranteed safe by type invariance, and we're guaranteed by DRM that
        // `self.state.crtc` points to a valid instance of a CRTC
        unsafe { <Self::Crtc as AsRawCrtc>::from_raw(self.as_raw().crtc) }
    }
}
impl<T: AsRawCrtcState> RawCrtcState for T {}

/// A trait for an object which can be derived directly from a raw DRM crtc state pointer.
pub trait FromRawCrtcState: AsRawCrtcState {
    /// Obtain a reference back to this type from a raw DRM crtc state pointer
    ///
    /// SAFETY: Callers must ensure that ptr contains a valid instance of this type..
    unsafe fn from_raw<'a>(ptr: *const bindings::drm_crtc_state) -> &'a Self;

    /// Obtain a mutable reference back to this type from a raw DRM crtc state pointer
    ///
    /// SAFETY: Callers must ensure that ptr contains a valid instance of this type, and that
    /// no other references to this CRTC's state exist for the lifetime of this reference
    unsafe fn from_raw_mut<'a>(ptr: *mut bindings::drm_crtc_state) -> &'a mut Self;
}

impl<T: DriverCrtcState> private::AsRawCrtcState for CrtcState<T> {
    #[inline]
    fn as_raw(&self) -> &bindings::drm_crtc_state {
        &self.state
    }

    #[inline]
    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_crtc_state {
        &mut self.state
    }
}

impl<T: DriverCrtcState> AsRawCrtcState for CrtcState<T> {
    type Crtc = Crtc<T::Crtc>;
}

impl<T: DriverCrtcState> FromRawCrtcState for CrtcState<T> {
    unsafe fn from_raw<'a>(ptr: *const bindings::drm_crtc_state) -> &'a Self {
        // SAFETY: Our data layout starts with `bindings::drm_crtc_state`
        unsafe { &*(ptr.cast()) }
    }

    unsafe fn from_raw_mut<'a>(ptr: *mut bindings::drm_crtc_state) -> &'a mut Self {
        // SAFETY: Our data layout starts with `bindings::drm_crtc_state`
        unsafe { &mut *(ptr.cast()) }
    }
}

/// A DRM crtc state without a known `DriverCrtcState` implementation
///
/// Used in situations where we can't automatically derive the `DriverCrtcState` implementation for
/// a CRTC state type.
#[repr(transparent)]
pub struct OpaqueCrtcState<T: KmsDriver> {
    state: bindings::drm_crtc_state,
    _p: PhantomData<T>
}

impl<T: KmsDriver> AsRawCrtcState for OpaqueCrtcState<T> {
    type Crtc = OpaqueCrtc<T>;
}

impl<T: KmsDriver> private::AsRawCrtcState for OpaqueCrtcState<T> {
    fn as_raw(&self) -> &bindings::drm_crtc_state {
        &self.state
    }

    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_crtc_state {
        &mut self.state
    }
}

impl<T: KmsDriver> FromRawCrtcState for OpaqueCrtcState<T> {
    unsafe fn from_raw<'a>(ptr: *const bindings::drm_crtc_state) -> &'a Self {
        // SAFETY: Our data layout is identical to `bindings::drm_crtc_state`
        unsafe { &*(ptr.cast()) }
    }

    unsafe fn from_raw_mut<'a>(ptr: *mut bindings::drm_crtc_state) -> &'a mut Self {
        // SAFETY: Our data layout is identical to `bindings::drm_crtc_state`
        unsafe { &mut *(ptr.cast()) }
    }
}

/// A mutable borrow from an `AtomicStateMutator` for a CRTC's atomic state
///
/// A user may use this object to access and/or mutate the driver's private CRTC state data, along
/// with modifying the CRTC's base atomic state. Since state mutators are unique in order to adhere
/// to rust's data aliasing rules, only one borrow may exist at a time for a CRTC's state.
pub struct BorrowedCrtcState<'a, T: FromRawCrtcState> {
    state: &'a mut T,
    mask: &'a Cell<u32>,
}

impl<'a, T: FromRawCrtcState> BorrowedCrtcState<'a, T> {
    pub(super) fn new<D: KmsDriver>(
        mutator: &'a AtomicStateMutator<D>,
        state: NonNull<bindings::drm_crtc_state>
    ) -> Option<Self> {
        // SAFETY: `crtc` is invariant throughout the lifetime of the atomic state, is
        // initialized by this point, and we're guaranteed it is of type `Crtc<T>` by type
        // invariance
        let crtc = unsafe { OpaqueCrtc::<D>::from_raw((*state.as_ptr()).crtc) };
        let crtc_mask = crtc.mask();
        let borrowed_mask = mutator.borrowed_crtcs.get();

        if borrowed_mask & crtc_mask == 0 {
            mutator.borrowed_crtcs.set(borrowed_mask | crtc_mask);
            Some(Self {
                mask: &mutator.borrowed_crtcs,
                // SAFETY: We're guaranteed `state` is of `CrtcState<T>` by type invariance, and we
                // just confirmed by checking `borrowed_crtcs` that no other mutable borrows have
                // been taken out for `state`
                state: unsafe { T::from_raw_mut(state.as_ptr()) },
            })
        } else {
            None
        }
    }
}

impl<'a, T: FromRawCrtcState> Drop for BorrowedCrtcState<'a, T> {
    fn drop(&mut self) {
        let mask = self.state.crtc().mask();
        self.mask.set(self.mask.get() & !mask);
    }
}

impl<'a, T: DriverCrtcState> Deref for BorrowedCrtcState<'a, CrtcState<T>> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.state.inner
    }
}

impl<'a, T: DriverCrtcState> DerefMut for BorrowedCrtcState<'a, CrtcState<T>> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.state.inner
    }
}

impl<'a, T: FromRawCrtcState> AsRawCrtcState for BorrowedCrtcState<'a, T> {
    type Crtc = T::Crtc;
}

impl<'a, T: FromRawCrtcState> private::AsRawCrtcState for BorrowedCrtcState<'a, T> {
    fn as_raw(&self) -> &bindings::drm_crtc_state {
        self.state.as_raw()
    }

    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_crtc_state {
        // SAFETY: This function is bound by the same safety contract as we are
        unsafe { self.state.as_raw_mut() }
    }
}

unsafe extern "C" fn crtc_destroy_callback<T: DriverCrtc>(
    crtc: *mut bindings::drm_crtc
) {
    // SAFETY: FFI call with no special requirements
    unsafe { bindings::drm_crtc_cleanup(crtc) };

    // SAFETY: We're guaranteed by type invariants this plane is contained within an Box<Crtc<T>>
    drop(unsafe { Box::from_raw(crtc as *mut Crtc<T>) });
}

unsafe extern "C" fn atomic_duplicate_state_callback<T: DriverCrtcState>(
    crtc: *mut bindings::drm_crtc
) -> *mut bindings::drm_crtc_state {
    // SAFETY: `crtc` has to point to a valid instance of drm_crtc, since it holds the vtable for
    // this function - which is the only possible entrypoint the caller could have used
    let state = unsafe { (*crtc).state };
    if state.is_null() {
        return null_mut();
    }

    // SAFETY: We just verified that `state` is non-null, and we're guaranteed by our bindings that
    // `state` is of type `CrtcState<T>`.
    let state = unsafe { CrtcState::<T>::from_raw(state) };

    let mut new: Result<Box<CrtcState<T>>> = Box::try_init(
        try_init!(CrtcState::<T> {
            state: bindings::drm_crtc_state { ..Default::default() },
            inner: state.inner.clone()
        }),
        GFP_KERNEL
    );

    if let Ok(mut new) = new {
        unsafe { bindings::__drm_atomic_helper_crtc_duplicate_state(crtc, new.as_raw_mut()) }

        Box::into_raw(new).cast()
    } else {
        null_mut()
    }
}

unsafe extern "C" fn atomic_destroy_state_callback<T: DriverCrtcState>(
    _crtc: *mut bindings::drm_crtc,
    crtc_state: *mut bindings::drm_crtc_state,
) {
    // SAFETY: This callback wouldn't be called unless there a CRTC state to destroy
    unsafe { bindings::__drm_atomic_helper_crtc_destroy_state(crtc_state) };
    //
    // SAFETY: We're guaranteed by type invariants that crtc_state is of type CrtcState<T>, and
    // since this is the destructor callback for DRM - we're guaranteed to hold the only remaining
    // reference to this state
    drop(unsafe { CrtcState::<T>::from_raw(crtc_state) });
}

unsafe extern "C" fn crtc_reset_callback<T: DriverCrtcState>(
    crtc: *mut bindings::drm_crtc,
) {
    // SAFETY: The only entrypoint to this function lives in `crtc` so it must be valid, and
    let state = unsafe { (*crtc).state };
    if !state.is_null() {
        // SAFETY: We're guaranteed by type invariance that this crtc's state is of type
        // DriverConnectorState<T>
        unsafe { atomic_destroy_state_callback::<T>(crtc, state) }

        // SAFETY: No special requirements here, DRM expects this to be NULL
        unsafe { (*crtc).state = null_mut(); }
    }

    // Unfortunately, this is the best we can do at the moment as this FFI callback was mistakenly
    // presumed to be infallible :(
    let new = Box::new(CrtcState::<T>::default(), GFP_KERNEL).expect("Blame the API, sorry!");

    // SAFETY: DRM takes ownership of the state from here and assigns it to the crtc
    unsafe { bindings::__drm_atomic_helper_crtc_reset(crtc, Box::into_raw(new).cast()) };
}

unsafe extern "C" fn atomic_check_callback<T: DriverCrtc>(
    crtc: *mut bindings::drm_crtc,
    state: *mut bindings::drm_atomic_state,
) -> i32 {
    // SAFETY: We're guaranteed `crtc` is of type `Crtc<T>` by type invariance, and we're guaranteed
    // it is pointing to a valid state by DRM
    let crtc = unsafe { Crtc::from_raw(crtc) };

    // SAFETY: Type conversion here is guaranteed safe by type invariance, DRM never passes an
    // invalid ptr for `state`
    let state = unsafe { LeakyAtomicStateComposer::new(NonNull::new_unchecked(state)) };

    // SAFETY: We're in the atomic update callback for this crtc, so we're guaranteed that both
    // crtc states are present in the atomic state already
    from_result(|| {
        T::atomic_check(crtc, &state)?;
        Ok(0)
    })
}
