// SPDX-License-Identifier: GPL-2.0 OR MIT

//! drm_encoder abstractions for rust

use crate::{
    drm::{
        device::{Device, PendingDevice},
        drv::Driver,
    },
    prelude::*,
    sync::Arc,
    types::Opaque,
    init::Zeroable,
    error::to_result,
    private::Sealed,
};
use core::{
    marker::*,
    ptr::{null, addr_of_mut},
    ops::Deref,
};
use super::{ModeObject, StaticModeObject, KmsDriver};
use bindings;

// TODO: We should alias all of these types with shorter names
pub use bindings::{
    DRM_MODE_ENCODER_NONE,
    DRM_MODE_ENCODER_DAC,
    DRM_MODE_ENCODER_TMDS,
    DRM_MODE_ENCODER_LVDS,
    DRM_MODE_ENCODER_TVDAC,
    DRM_MODE_ENCODER_VIRTUAL,
    DRM_MODE_ENCODER_DSI,
    DRM_MODE_ENCODER_DPMST,
    DRM_MODE_ENCODER_DPI,
};

/// The main trait for KMS drivers to implement for their display encoders.
pub trait DriverEncoder: Send + Sync + Sized {
    /// The parent driver for this drm_encoder implementation
    type Driver: KmsDriver;

    /// The type used for passing arguments to the driver's constructor
    /// Drivers which don't care about this can just use `()`
    type Args;

    /// Create a new encoder for this driver
    fn new(device: &Device<Self::Driver>, args: Self::Args) -> impl PinInit<Self, Error>;
}

/// A trait for objects which can be used as a DRM encoder.
pub trait AsRawEncoder: StaticModeObject {
    /// Return the raw `bindings::drm_encoder` for this DRM encoder.
    ///
    /// Drivers should never use this directly
    fn as_raw(&self) -> *mut bindings::drm_encoder;

    /// Convert a raw `bindings::drm_encoder` pointer into an object of this type.
    ///
    /// SAFETY: Callers promise that `ptr` points to a valid instance of this type
    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_encoder) -> &'a Self;
}

/// A typed DRM encoder (`drm_encoder`)
///
/// This is the main struct for DRM encoders, which may also hold any private data specified by the
/// driver.
#[repr(C)]
#[pin_data]
pub struct Encoder<T: DriverEncoder> {
    /// The FFI drm_encoder object
    encoder: Opaque<bindings::drm_encoder>,
    /// The driver's private inner data
    #[pin]
    inner: T,
    #[pin]
    _p: PhantomPinned,
}

impl<T: DriverEncoder> Sealed for Encoder<T> {}

unsafe impl Zeroable for bindings::drm_encoder {}

// SAFETY: Encoder.encoder is not exposed to users by default, and our accessors ensure we only
// perform thread-safe operations for this object
unsafe impl<T: DriverEncoder> Send for Encoder<T> { }
unsafe impl<T: DriverEncoder> Sync for Encoder<T> { }

impl<T: DriverEncoder> ModeObject for Encoder<T> {
    type Driver = T::Driver;

    fn drm_dev(&self) -> &Device<Self::Driver> {
        // SAFETY: DRM encoders exist for as long as the device does, so this pointer is always
        // valid
        unsafe { Device::borrow((*self.encoder.get()).dev) }
    }

    fn raw_mode_obj(&self) -> *mut bindings::drm_mode_object {
        // SAFETY: We don't expose Encoder<T> to users before it's initialized, so `base` is always
        // initialized
        unsafe { &mut (*self.encoder.get()).base }
    }
}

// SAFETY: Encoders do not have a refcount
unsafe impl<T: DriverEncoder> StaticModeObject for Encoder<T> { }

impl<T: DriverEncoder> Deref for Encoder<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl<T: DriverEncoder> AsRawEncoder for Encoder<T> {
    fn as_raw(&self) -> *mut bindings::drm_encoder {
        self.encoder.get()
    }

    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_encoder) -> &'a Self {
        // SAFETY: Our data layout is starts with to `bindings::drm_encoder`
        unsafe { &*ptr.cast() }
    }
}

impl<T: DriverEncoder> Encoder<T> {
    const FUNCS: bindings::drm_encoder_funcs = bindings::drm_encoder_funcs {
        reset: None,
        destroy: Some(encoder_destroy_callback::<T>),
        late_register: None,
        early_unregister: None,
        debugfs_init: None,
    };

    const HELPER_FUNCS: bindings::drm_encoder_helper_funcs = bindings::drm_encoder_helper_funcs {
        dpms: None,
        mode_valid: None,
        mode_fixup: None,
        prepare: None,
        mode_set: None,
        commit: None,
        detect: None,
        enable: None,
        disable: None,
        atomic_check: None,
        atomic_enable: None,
        atomic_disable: None,
        atomic_mode_set: None,
    };

    /// Create a new encoder for this unregistered DRM device
    pub fn new<'a, 'b: 'a>(
        dev: &'a PendingDevice<T::Driver>,
        type_: u32,
        possible_crtcs: u32,
        possible_clones: u32,
        name: Option<&CStr>,
        args: T::Args,
    ) -> Result<&'b Self> {
        let this: Pin<Box<Self>> = Box::try_pin_init(
            try_pin_init!(Self {
                encoder: Opaque::new(bindings::drm_encoder {
                    helper_private: &Self::HELPER_FUNCS,
                    possible_crtcs,
                    possible_clones,
                    ..Default::default()
                }),
                inner <- T::new(dev, args),
                _p: PhantomPinned
            }),
            GFP_KERNEL
        )?;

        // SAFETY: FFI call with no special requirements
        to_result(unsafe {
            bindings::drm_encoder_init(
                dev.as_raw(),
                this.as_raw(),
                &Self::FUNCS,
                type_ as _,
                name.map_or(null(), |n| n.as_char_ptr())
            )
        })?;

        // Convert the box into a raw pointer, we'll re-assemble it in encoder_destroy_callback()
        // SAFETY: We don't move anything
        Ok(unsafe { &*Box::into_raw(Pin::into_inner_unchecked(this)) })
    }
}

#[repr(transparent)]
pub struct OpaqueEncoder<T: KmsDriver> {
    encoder: Opaque<bindings::drm_encoder>,
    _p: PhantomData<T>,
}

impl<T: KmsDriver> Sealed for OpaqueEncoder<T> {}

// SAFETY: All of our encoder interfaces are thread-safe
unsafe impl<T: KmsDriver> Send for OpaqueEncoder<T> {}
unsafe impl<T: KmsDriver> Sync for OpaqueEncoder<T> {}

impl<T: KmsDriver> ModeObject for OpaqueEncoder<T> {
    type Driver = T;

    fn drm_dev(&self) -> &Device<Self::Driver> {
        // SAFETY: DRM encoders exist for as long as the device does, so this pointer is always
        // valid
        unsafe { Device::borrow((*self.encoder.get()).dev) }
    }

    fn raw_mode_obj(&self) -> *mut bindings::drm_mode_object {
        // SAFETY: We don't expose Encoder<T> to users before it's initialized, so `base` is always
        // initialized
        unsafe { addr_of_mut!((*self.encoder.get()).base) }
    }
}

// SAFETY: Encoders do not have a refcount
unsafe impl<T: KmsDriver> StaticModeObject for OpaqueEncoder<T> {}

impl<T: KmsDriver> AsRawEncoder for OpaqueEncoder<T> {
    fn as_raw(&self) -> *mut bindings::drm_encoder {
        self.encoder.get()
    }

    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_encoder) -> &'a Self {
        // SAFETY: Our data layout is identical to `bindings::drm_encoder`
        unsafe { &*ptr.cast() }
    }
}

unsafe extern "C" fn encoder_destroy_callback<T: DriverEncoder>(
    encoder: *mut bindings::drm_encoder
) {
    // SAFETY: encoder contains the only possible entrypoint to this function, so the pointer must
    // be valid
    unsafe { bindings::drm_encoder_cleanup(encoder) };

    // Reclaim ownership of the reference we took in Encoder::<T>::new() so we can drop it
    // SAFETY: We always create encoders in Arc<T>, and we're guaranteed by type invariants that
    // this encoder is a Encoder<T>
    unsafe { drop(Box::from_raw(encoder as *mut Encoder<T>)) };
}
