use crate::{
    drm::{
        device::Device,
        fourcc::*,
    },
    types::*,
    prelude::*,
};
use bindings;
use core::{
    marker::*,
    slice,
};
use super::{ModeObject, RcModeObject, KmsDriver};

#[repr(transparent)]
pub struct Framebuffer<T: KmsDriver> {
    inner: Opaque<bindings::drm_framebuffer>,
    _p: PhantomData<T>,
}

impl<T: KmsDriver> ModeObject for Framebuffer<T> {
    type Driver = T;

    fn drm_dev(&self) -> &Device<Self::Driver> {
        unsafe { Device::borrow((*self.inner.get()).dev) }
    }

    fn raw_mode_obj(&self) -> *mut bindings::drm_mode_object {
        // SAFETY: We don't expose Framebuffer<T> to users before it's initialized, so `base` is
        // always initialized
        unsafe { &mut (*self.inner.get()).base }
    }
}

// SAFETY: Framebuffers are refcounted mode objects.
unsafe impl<T: KmsDriver> RcModeObject for Framebuffer<T> {}

// SAFETY: References to framebuffers are safe to be accessed from any thread
unsafe impl<T: KmsDriver> Send for Framebuffer<T> {}
unsafe impl<T: KmsDriver> Sync for Framebuffer<T> {}

impl<T: KmsDriver> crate::private::Sealed for Framebuffer<T> {}

impl<T: KmsDriver> Framebuffer<T> {
    #[inline]
    pub(super) unsafe fn from_raw<'a>(ptr: *const bindings::drm_framebuffer) -> &'a Self {
        // SAFETY: Our data layout is identical to drm_framebuffer
        unsafe { &*ptr.cast() }
    }

    /// Get the offset from buffer start to the actual pixel data in bytes, per buffer.
    pub fn offsets(&self) -> &[u32] {
        // SAFETY: `offsets` is initialized as long as we expose `Framebuffer<T>` to users
        unsafe { &(*self.inner.get()).offsets[..self.format().num_components() as _] }
    }

    /// Get the line stride per buffer.
    pub fn pitches(&self) -> &[u32] {
        // SAFETY: `pitches` is initialized as long as we expose `Framebuffer<T>` to users
        unsafe { &(*self.inner.get()).pitches[..self.format().num_components() as _] }
    }

    /// Get the format information for this framebuffer
    #[inline]
    pub fn format(&self) -> &FormatInfo {
        // SAFETY: `format` is initialized as long as we expose `Framebuffer<T>` to users
        unsafe { FormatInfo::from_raw((*self.inner.get()).format) }
    }
}
