//! DRM's atomic GEM helpers

use crate::{
    error::to_result,
    prelude::*,
    private::Sealed,
    bindings,
    init::Zeroable,
    container_of,
};
use core::{
    ops::*,
    mem::{self, offset_of},
};
use super::{
    atomic::*,
    plane::*,
    crtc::*,
};

unsafe impl Zeroable for bindings::drm_shadow_plane_state {}

/// A plane state which uses DRM's shadow plane state helpers
///
/// Drivers may use this instead of `PlaneState<T>` in their `DriverPlaneState`
/// implementation to access DRM's shadow plane helper.
#[derive(Default)]
#[repr(C)]
pub struct ShadowPlaneState<T: DriverPlaneState> {
    state: bindings::drm_shadow_plane_state,
    inner: T,
}

impl<T: DriverPlaneState> Sealed for ShadowPlaneState<T> {}

// A lot of code here makes the assumption that the base `bindings::drm_plane_state` is at the
// beginning of the struct. That's very unlikely to change, but better safe then sorry.
static_assert!(offset_of!(bindings::drm_shadow_plane_state, base) == 0);

impl<T: DriverPlaneState> AsRawPlaneState for ShadowPlaneState<T> {
    type Driver = <T::Plane as DriverPlane>::Driver;
    type Plane = Plane<T::Plane>;
}

impl<T: DriverPlaneState> AsRawPlaneStatePrivate for ShadowPlaneState<T> {
    #[inline]
    fn as_raw(&self) -> &bindings::drm_plane_state {
        &self.state.base
    }

    #[inline]
    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_plane_state {
        &mut self.state.base
    }
}

impl<T: DriverPlaneState> FromRawPlaneState for ShadowPlaneState<T> {
    #[inline]
    unsafe fn from_raw<'a>(ptr: *const bindings::drm_plane_state) -> &'a Self {
        // SAFETY: Our data layout starts with `bindings::drm_plane_state`
        unsafe { &*ptr.cast() }
    }

    #[inline]
    unsafe fn from_raw_mut<'a>(ptr: *mut bindings::drm_plane_state) -> &'a mut Self {
        // SAFETY: Our data layout starts with `bindings::drm_plane_state`
        unsafe { &mut *ptr.cast() }
    }
}

impl<T: DriverPlaneState> PlaneStateHelper for ShadowPlaneState<T> {
    type DriverState = T;

    fn __duplicate_state(&self, plane: *mut bindings::drm_plane) -> Result<Box<Self>> {
        let mut new: Box<Self> = Box::try_init(
            try_init!(Self {
                state: bindings::drm_shadow_plane_state { ..Default::default() },
                inner: self.inner.clone()
            }),
            GFP_KERNEL
        )?;

        // SAFETY: FFI call with no special requirements
        unsafe { bindings::__drm_gem_duplicate_shadow_plane_state(plane, &mut new.state) };

        Ok(new)
    }

    fn __destroy_state(state: *mut bindings::drm_plane_state) {
        // SAFETY: This would not be called without a plane state to destroy, and our data layout
        // starts with `bindings::drm_plane_state`
        unsafe { bindings::__drm_gem_destroy_shadow_plane_state(state.cast()) };
    }

    fn __reset_state(plane: *mut bindings::drm_plane, state: Box<Self>) {
        // SAFETY: This would not be called without a plane state to reset, and our data layout
        // starts with `bindings::drm_plane_state`
        unsafe { bindings::__drm_gem_reset_shadow_plane(plane, Box::into_raw(state).cast()) }
    }

    #[inline]
    fn inner(&self) -> &Self::DriverState {
        &self.inner
    }

    #[inline]
    fn inner_mut(&mut self) -> &mut Self::DriverState {
        &mut self.inner
    }
}

impl<T: DriverPlaneState> Deref for ShadowPlaneState<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl<T: DriverPlaneState> DerefMut for ShadowPlaneState<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.inner
    }
}

pub trait RawShadowPlaneState: AsRawShadowPlaneState {
    fn gem_plane_helper_prepare_fb(&mut self) -> Result {
        // SAFETY: FFI call with no special safety requirements
        to_result(unsafe {
            bindings::drm_gem_plane_helper_prepare_fb(
                self.plane().as_raw(),
                <Self as AsRawPlaneStatePrivate>::as_raw_mut(self)
            )
        })
    }

    fn gem_plane_helper_begin_fb_access(&mut self) -> Result {
        // SAFETY: FFI call with no special safety requirements
        to_result(unsafe {
            bindings::drm_gem_begin_shadow_fb_access(
                self.plane().as_raw(),
                <Self as AsRawPlaneStatePrivate>::as_raw_mut(self)
            )
        })
    }

    fn gem_plane_helper_end_fb_access(&mut self) {
        // SAFETY: FFI call with no special safety requirements
        unsafe {
            bindings::drm_gem_end_shadow_fb_access(
                self.plane().as_raw(),
                <Self as AsRawPlaneStatePrivate>::as_raw_mut(self)
            )
        }
    }
}
impl<T: AsRawShadowPlaneState> RawShadowPlaneState for T {}

pub(crate) mod private {
    use super::*;

    pub trait AsRawShadowPlaneState: AsRawPlaneState {
        /// Return an immutable reference to the base `bindings::drm_shadow_plane_state`
        fn as_raw(&self) -> &bindings::drm_shadow_plane_state;

        /// Get a mutable reference to the raw `bindings::drm_shadow_plane_state` contained within
        /// this type.
        ///
        /// SAFETY: The caller promises this mutable reference will not be used to modify any
        /// contents of `bindings::drm_shadow_plane_state` which DRM would consider to be static -
        /// like the backpointer to the DRM plane that owns this state. This also means the mutable
        /// reference should never be exposed outside of this crate.
        unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_shadow_plane_state;
    }
}

pub trait AsRawShadowPlaneState: private::AsRawShadowPlaneState {}

impl<T, S, P> AsRawShadowPlaneState for S
where
    T: DriverPlaneState,
    S: AsRawPlaneState<Plane = Plane<P>>,
    P: DriverPlane<State = ShadowPlaneState<T>>
{}

impl<T, S, P> private::AsRawShadowPlaneState for S
where
    T: DriverPlaneState,
    S: AsRawPlaneState<Plane = Plane<P>>,
    P: DriverPlane<State = ShadowPlaneState<T>>
{
    fn as_raw(&self) -> &bindings::drm_shadow_plane_state {
        // SAFETY: Our data layout starts with `bindings::drm_plane_state`
        unsafe { mem::transmute(AsRawPlaneStatePrivate::as_raw(self)) }
    }

    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_shadow_plane_state {
        // SAFETY: Our data layout starts with `bindings::drm_plane_state`
        unsafe { mem::transmute(AsRawPlaneStatePrivate::as_raw_mut(self)) }
    }
}
