// SPDX-License-Identifier: GPL-2.0 OR MIT

//! KMS atomic plane abstractions for rust.

use alloc::boxed::Box;
use crate::{
    bindings,
    drm::{device::{Device, PendingDevice}, drv::Driver, rect::Rect},
    error::{to_result, from_err_ptr, from_result, Error},
    init::Zeroable,
    prelude::*,
    types::{ARef, Opaque},
    private::Sealed,
};
use core::{
    cell::Cell,
    pin::Pin,
    slice,
    mem::{self, size_of, MaybeUninit},
    ptr::{self, null, null_mut, NonNull},
    marker::*,
    ops::*,
    ptr::{drop_in_place, addr_of_mut},
};
use macros::pin_data;
use super::{
    KmsDriver,
    ModeObject,
    StaticModeObject,
    atomic::*,
    crtc::*,
    framebuffer::*,
};

/// The main trait for implementing the drm_plane API. This contains the various trait methods that
/// need to be implemented by a driver. The private driver data for the plane is contained in
/// whatever struct the driver defines which implements this trait.
#[vtable]
pub trait DriverPlane: Send + Sync + Sized {
    /// The data type to use for passing incoming arguments for new `Plane<T>` instances
    /// Drivers which don't care about this can just use `()`
    type Args;

    /// The parent driver implementation
    type Driver: KmsDriver;

    /// The full type for this driver's drm_plane_state implementation. Drivers which don't need
    /// special DRM helpers for their plane states may just use `PlaneState<T>` here, where `T` is
    /// their private state struct which implements `DriverPlaneState`
    type State: PlaneStateHelper;

    /// Create the plane driver's private data structure
    fn new(device: &Device<Self::Driver>, args: Self::Args) -> impl PinInit<Self, Error>;

    /// The atomic check hook
    fn atomic_check(plane: &Plane<Self>, state: &AtomicStateComposer<Self::Driver>) -> Result {
        build_error::build_error("This should not be reachable")
    }

    /// The atomic_update hook for this plane.
    ///
    /// XXX: Currently not optional, but should be in the future once we add support for
    /// atomic_enable/disable
    fn atomic_update(plane: &Plane<Self>, state: &AtomicStateMutator<Self::Driver>);

    /// Callback for any work the driver needs to perform for preparing a framebuffer
    fn prepare_fb(plane: &Plane<Self>, state: &mut Self::State) -> Result {
        build_error::build_error("This should not be reachable")
    }

    /// Callback for any work the driver needs to perform for cleaning up a framebuffer that's no
    /// longer in use. This should be the inverse of `prepare_fb()`
    fn cleanup_fb(plane: &Plane<Self>, state: &mut Self::State) {
        build_error::build_error("This should not be reachable")
    }
}

/// The main structure containing a drm_plane that is exposed to callers. It is intentionally
/// impossible to acquire a mutable reference to this structure, and as such this structure should
/// only be exposed through immutable references.
#[repr(C)]
#[pin_data]
pub struct Plane<T: DriverPlane> {
    /// The FFI drm_plane object
    plane: Opaque<bindings::drm_plane>,
    /// The driver's private inner data
    #[pin]
    inner: T,
    #[pin]
    _p: PhantomPinned,
}

unsafe impl Zeroable for bindings::drm_plane {}

impl<T: DriverPlane> Sealed for Plane<T> {}

impl<T: DriverPlane> Deref for Plane<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl<T: DriverPlane> Plane<T> {
    /// The actual C vtable for drm_plane_funcs
    const FUNCS: bindings::drm_plane_funcs = bindings::drm_plane_funcs {
        update_plane: Some(bindings::drm_atomic_helper_update_plane),
        disable_plane: Some(bindings::drm_atomic_helper_disable_plane),
        destroy: Some(plane_destroy_callback::<T>),
        reset: Some(plane_reset_callback::<T::State>),
        set_property: None,
        atomic_duplicate_state: Some(atomic_duplicate_state_callback::<T::State>),
        atomic_destroy_state: Some(atomic_destroy_state_callback::<T::State>),
        atomic_set_property: None, // TODO someday
        atomic_get_property: None, // TODO someday
        late_register: None, // TODO someday
        early_unregister: None, // TODO someday
        atomic_print_state: None, // TODO: Display someday???
        format_mod_supported: None // TODO someday
    };

    /// The actual C vtable for drm_plane_helper_funcs
    const HELPER_FUNCS: bindings::drm_plane_helper_funcs = bindings::drm_plane_helper_funcs {
        prepare_fb: if T::HAS_PREPARE_FB { Some(prepare_fb_callback::<T>) } else { None },
        cleanup_fb: if T::HAS_CLEANUP_FB { Some(cleanup_fb_callback::<T>) } else { None },
        begin_fb_access: None, // TODO: someday?
        end_fb_access: None, // TODO: someday?
        atomic_check: if T::HAS_ATOMIC_CHECK { Some(atomic_check_callback::<T>) } else { None },
        atomic_update: Some(atomic_update_callback::<T>),
        atomic_enable: None, // TODO
        atomic_disable: None, // TODO
        atomic_async_check: None, // TODO
        atomic_async_update: None, // TODO
        panic_flush: None,
        get_scanout_buffer: None
    };

    /// Create a new DRM plane for `dev`
    pub fn new<'a, 'b: 'a>(
        dev: &'a PendingDevice<T::Driver>,
        possible_crtcs: u32,
        formats: &'static [u32],
        format_modifiers: Option<&'static [u64]>,
        type_: PlaneType,
        name: Option<&CStr>,
        args: T::Args,
    ) -> Result<&'b Self> {
        let this: Pin<Box<Self>> = Box::try_pin_init(
            try_pin_init!(Self {
                plane: Opaque::new(bindings::drm_plane {
                    helper_private: &Self::HELPER_FUNCS,
                    ..Default::default()
                }),
                inner <- T::new(dev, args),
                _p: PhantomPinned
            }),
            GFP_KERNEL
        )?;

        // SAFETY: FFI call with no special requirements
        to_result(unsafe {
            bindings::drm_universal_plane_init(
                dev.as_raw(),
                this.as_raw(),
                possible_crtcs,
                &Self::FUNCS,
                formats.as_ptr(),
                formats.len() as _,
                format_modifiers.map_or(null(), |f| f.as_ptr()),
                type_ as _,
                name.map_or(null(), |n| n.as_char_ptr())
            )
        })?;

        // Convert the box into a raw pointer, we'll re-assemble it in plane_destroy_callback()
        // SAFETY: We don't move anything
        Ok(unsafe { &*Box::into_raw(Pin::into_inner_unchecked(this)) })
    }

    /// Attempt to convert an `OpaquePlane<D>` into this type.
    ///
    /// If the underlying DRM plane uses a different `DriverPlane` implementation, `None` is
    /// returned. The plane must be from the same `KmsDriver` implementation as the opaque plane.
    pub fn from_opaque<'a, D>(opaque: &'a OpaquePlane<D>) -> Option<&'a Self>
    where
        D: KmsDriver,
        T: DriverPlane<Driver = D>
    {
        // SAFETY: The vtables for a `Plane` are initialized by the time that we expose `Plane`
        // objects to users, and their values are invariant throughout the lifetime of the device.
        let (funcs, helper_funcs) = unsafe {
            let plane = opaque.plane.get();

            ((*plane).funcs, (*plane).helper_private)
        };

        // SAFETY: We just guaranteed that the opaque plane shares our vtable pointers, which means
        // it must belong to our `DriverPlane` implementation. As well, all `Plane<DriverPlane>`
        // objects start with an identical data layout to `OpaquePlane`
        (ptr::eq(funcs, &Self::FUNCS) && ptr::eq(helper_funcs, &Self::HELPER_FUNCS))
            .then(|| unsafe { mem::transmute(opaque) })
    }
}

/// A trait for objects which can be used as a DRM plane.
pub trait AsRawPlane: StaticModeObject {
    /// The type that should be returned for a plane state acquired using this plane interface
    type State: AsRawPlaneState;

    /// Return the raw `bindings::drm_plane` for this DRM plane.
    ///
    /// Drivers should never use this directly
    fn as_raw(&self) -> *mut bindings::drm_plane;

    /// Convert a raw `bindings::drm_plane` pointer into an object of this type.
    ///
    /// SAFETY: Callers promise that `ptr` points to a valid instance of this type
    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_plane) -> &'a Self;
}

impl<T: DriverPlane> AsRawPlane for Plane<T> {
    type State = T::State;

    fn as_raw(&self) -> *mut bindings::drm_plane {
        self.plane.get()
    }

    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_plane) -> &'a Self {
        // SAFETY: Our data layout starts with `bindings::drm_plane`
        unsafe { &*ptr.cast() }
    }
}

impl<T: DriverPlane> ModeObject for Plane<T> {
    type Driver = T::Driver;

    fn drm_dev(&self) -> &Device<Self::Driver> {
        // SAFETY: DRM planes exist for as long as the device does, so this pointer is always valid
        unsafe { Device::borrow((*self.as_raw()).dev) }
    }

    fn raw_mode_obj(&self) -> *mut bindings::drm_mode_object {
        // SAFETY: We don't expose DRM planes to users before `base` is initialized
        unsafe { addr_of_mut!((*self.as_raw()).base) }
    }
}

// SAFETY: Planes do not have a refcount
unsafe impl<T: DriverPlane> StaticModeObject for Plane<T> {}

// SAFETY: Our plane interfaces are guaranteed to be thread-safe
unsafe impl<T: DriverPlane> Send for Plane<T> {}
unsafe impl<T: DriverPlane> Sync for Plane<T> {}

/// Common methods available on any type which can be used as a DRM plane
pub trait RawPlane: AsRawPlane {
    /// Return the index of this DRM plane
    #[inline]
    fn index(&self) -> u32 {
        // SAFETY: The index is initialized by the time we expose `Plane` objects to users, and is
        // invariant throughout the lifetime of the `Plane`
        unsafe { (*self.as_raw()).index }
    }

    /// Return the index of this DRM plane in the form of a bitmask
    #[inline]
    fn mask(&self) -> u32 {
        1 << self.index()
    }
}
impl<T: AsRawPlane> RawPlane for T {}

/// A DRM plane without a known `PlaneDriver` implementation
///
/// Mainly for situations where our bindings can't infer the `PlaneDriver` implementation
/// automatically. Drivers can use `Plane::<PlaneDriver>::from_opaque` to attempt upcast the opaque
/// plane at runtime.
#[repr(transparent)]
pub struct OpaquePlane<T: KmsDriver> {
    plane: Opaque<bindings::drm_plane>,
    _p: PhantomData<T>,
}

impl<T: KmsDriver> Sealed for OpaquePlane<T> {}

impl<T: KmsDriver> AsRawPlane for OpaquePlane<T> {
    type State = OpaquePlaneState<T>;

    fn as_raw(&self) -> *mut bindings::drm_plane {
        self.plane.get()
    }

    unsafe fn from_raw<'a>(ptr: *mut bindings::drm_plane) -> &'a Self {
        // SAFETY: Our data layout is identical to `bindings::drm_plane`
        unsafe { &*ptr.cast() }
    }
}

impl<T: KmsDriver> ModeObject for OpaquePlane<T> {
    type Driver = T;

    fn drm_dev(&self) -> &Device<Self::Driver> {
        // SAFETY: DRM planes exist for as long as the device does, so this pointer is always valid
        unsafe { Device::borrow((*self.as_raw()).dev) }
    }

    fn raw_mode_obj(&self) -> *mut bindings::drm_mode_object {
        // SAFETY: We don't expose DRM planes to users before `base` is initialized
        unsafe { addr_of_mut!((*self.as_raw()).base) }
    }
}

// SAFETY: Planes do not have a refcount
unsafe impl<T: KmsDriver> StaticModeObject for OpaquePlane<T> {}

// SAFETY: Our plane interfaces are guaranteed to be thread-safe
unsafe impl<T: KmsDriver> Send for OpaquePlane<T> {}
unsafe impl<T: KmsDriver> Sync for OpaquePlane<T> {}

/// A trait for any type that can be used as the main wrapper around the bindings::drm_plane_state
/// for a plane.
/// TODO: Make all of the inner members here private to the crate
pub trait PlaneStateHelper: Sized + Default + FromRawPlaneState {
    /// The type of the private driver state data contained within this helper
    type DriverState: DriverPlaneState;

    /// Allocate a "duplicated" copy of this state, which usually involves calling DRM's helper
    /// function for this plane state type
    fn __duplicate_state(&self, plane: *mut bindings::drm_plane) -> Result<Box<Self>>;

    /// Call DRM's destroy helper for this type of plane state. Note this only cleans up the base DRM
    /// state struct and does not de-allocate its `Box`.
    fn __destroy_state(state: *mut bindings::drm_plane_state);

    /// Call DRM's reset helper for this type of plane state.
    fn __reset_state(plane: *mut bindings::drm_plane, state: Box<Self>);

    /// Return an immutable reference to the driver's private plane state
    fn inner(&self) -> &Self::DriverState;

    /// Return a mutable reference to the driver's private plane state
    fn inner_mut(&mut self) -> &mut Self::DriverState;
}

/// A type which which can produce a reference to a raw DRM plane state structure.
///
/// This can be any type that contains a `bindings::drm_plane_state`, or contains a reference to a
/// `bindings::drm_plane_state`. Or in other words - any rust type that can be used as a DRM plane
/// state. Note that unlike `AsRawPlane`, not all objects which implement `AsRawPlaneState` can be
/// obtained from raw `bindings::drm_plane_state` pointers.
pub trait AsRawPlaneState: private::AsRawPlaneState {
    /// The parent driver for this DRM CRTC interface
    type Driver: KmsDriver;

    /// The type that this plane state interface returns to represent the parent DRM plane
    type Plane: AsRawPlane;
}

pub(crate) mod private {
    /// Trait for retrieving references to the base plane state contained within any plane state
    /// compatible type
    #[doc(hidden)]
    pub trait AsRawPlaneState {
        /// Return an immutable reference to the raw plane state
        fn as_raw(&self) -> &bindings::drm_plane_state;

        /// Get a mutable reference to the raw `bindings::drm_plane_state` contained within this
        /// type.
        ///
        /// SAFETY: The caller promises this mutable reference will not be used to modify any
        /// contents of `bindings::drm_plane_state` which DRM would consider to be static - like the
        /// backpointer to the DRM plane that owns this state. This also means the mutable reference
        /// should never be exposed outside of this crate.
        unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_plane_state;
    }
}

pub(crate) use private::AsRawPlaneState as AsRawPlaneStatePrivate;

/// A type which can be obtained directly from a raw DRM plane state pointer.
///
/// This can be any type which can be created directly from a raw pointer to a
/// `bindings::drm_plane_state`, and which can be used as a plane state.
pub trait FromRawPlaneState: AsRawPlaneState {
    /// Get an immutable reference to this type from the given raw `bindings::drm_plane_state`
    /// pointer
    ///
    /// SAFETY: The caller guarantees `ptr` is contained within a valid instance of `Self`
    unsafe fn from_raw<'a>(ptr: *const bindings::drm_plane_state) -> &'a Self;

    /// Get a mutable reference to this type from the given raw `bindings::drm_plane_state` pointer
    ///
    /// SAFETY: The caller guarantees `ptr` is contained within a valid instance of `Self`, and that
    /// no other references (mutable or immutable) to `ptr` exist.
    unsafe fn from_raw_mut<'a>(ptr: *mut bindings::drm_plane_state) -> &'a mut Self;
}

/// Common methods available on any type provided by DRM that can act as an atomic plane state. They
/// cannot be overriden.
pub trait RawPlaneState: AsRawPlaneState {
    /// Return the plane that this state belongs to
    fn plane(&self) -> &Self::Plane {
        // SAFETY: The index is initialized by the time we expose Plane objects to users, and is
        // invariant throughout the lifetime of the Plane
        unsafe { Self::Plane::from_raw(self.as_raw().plane) }
    }

    /// Return the current CRTC assigned to this plane, if any.
    fn crtc<'a, 'b: 'a, D, P>(&'a self) -> Option<&'b OpaqueCrtc<D>>
    where
        Self: AsRawPlaneState<Plane = P>,
        P: ModeObject<Driver = D>,
        D: KmsDriver
    {
        // SAFETY: We're guaranteed `crtc` is of type `OpaqueCrtc` by type invariance
        NonNull::new(self.as_raw().crtc).map(|c| unsafe { OpaqueCrtc::from_raw(c.as_ptr()) })
    }

    /// Run the atomic check helper for this plane and the given CRTC state
    fn atomic_helper_check<C: FromRawCrtcState>(
        &mut self,
        crtc_state: &BorrowedCrtcState<'_, C>,
        can_position: bool,
        can_update_disabled: bool
    ) -> Result {
        // SAFETY: We're passing the mutable reference from `self.as_raw_mut()` directly to DRM,
        // which is safe.
        to_result(unsafe {
            bindings::drm_atomic_helper_check_plane_state(
                self.as_raw_mut(),
                crtc_state.as_raw(),
                bindings::DRM_PLANE_NO_SCALING as _, // TODO: add parameters for scaling
                bindings::DRM_PLANE_NO_SCALING as _,
                can_position,
                can_update_disabled
            )
        })
    }

    /// Return the framebuffer currently set for this plane state
    #[inline]
    fn framebuffer(&self) -> Option<&Framebuffer<Self::Driver>> {
        // SAFETY: The layout of Framebuffer<T> is identical to `fb`
        unsafe { self.as_raw().fb.as_ref().map(|fb| Framebuffer::from_raw(fb)) }
    }

    /// The (maybe clipped) source coordinates of the plane
    ///
    /// When using `Plane::atomic_helper_check()`, the coordinates are clipped, but the driver may
    /// choose to use unclipped coordinates instead when the hardware performs the clipping
    /// automatically.
    #[inline]
    fn src(&self) -> &Rect {
        self.as_raw().src.as_ref()
    }

    /// The (maybe clipped) dst coordinates of the plane
    ///
    /// When using `Plane::atomic_helper_check()`, the coordinates are clipped, but the driver may
    /// choose to use unclipped coordinates instead when the hardware performs the clipping
    /// automatically.
    #[inline]
    fn dst(&self) -> &Rect {
        self.as_raw().dst.as_ref()
    }

    /// The rotation of the plane
    #[inline]
    fn rotation(&self) -> u32 {
        self.as_raw().rotation
    }
}
impl<T: AsRawPlaneState + ?Sized> RawPlaneState for T {}

#[derive(Default)]
#[repr(C)]
pub struct PlaneState<T: DriverPlaneState> {
    state: bindings::drm_plane_state,
    inner: T,
}

/// Traits which must be implemented by KMS drivers for DRM planes.
pub trait DriverPlaneState: Clone + Default + Sized {
    /// The type for this driver's drm_plane implementation
    type Plane: DriverPlane;
}

impl<T: DriverPlaneState> Sealed for PlaneState<T> {}

impl<T: DriverPlaneState> AsRawPlaneState for PlaneState<T> {
    type Driver = <T::Plane as DriverPlane>::Driver;
    type Plane = Plane<T::Plane>;
}

impl<T: DriverPlaneState> private::AsRawPlaneState for PlaneState<T> {
    fn as_raw(&self) -> &bindings::drm_plane_state {
        &self.state
    }

    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_plane_state {
        &mut self.state
    }
}

impl<T: DriverPlaneState> FromRawPlaneState for PlaneState<T> {
    unsafe fn from_raw<'a>(ptr: *const bindings::drm_plane_state) -> &'a Self {
        // SAFETY: Our data layout starts with `bindings::drm_plane_state`
        unsafe { &*ptr.cast() }
    }

    unsafe fn from_raw_mut<'a>(ptr: *mut bindings::drm_plane_state) -> &'a mut Self {
        // SAFETY: Our data layout starts with `bindings::drm_plane_state`
        unsafe { &mut *ptr.cast() }
    }
}

impl<T: DriverPlaneState> PlaneStateHelper for PlaneState<T> {
    type DriverState = T;

    fn __duplicate_state(&self, plane: *mut bindings::drm_plane) -> Result<Box<Self>> {
        let mut new: Box<Self> = Box::try_init(
            try_init!(Self {
                state: bindings::drm_plane_state { ..Default::default() },
                inner: self.inner.clone(),
            }),
            GFP_KERNEL
        )?;

        // SAFETY: We're passing `new.as_raw_mut()` directly to DRM, which is safe
        unsafe { bindings::__drm_atomic_helper_plane_duplicate_state(plane, new.as_raw_mut()) };

        Ok(new)
    }

    fn __destroy_state(state: *mut bindings::drm_plane_state) {
        // SAFETY: This would not be called without a plane state to destroy
        unsafe { bindings::__drm_atomic_helper_plane_destroy_state(state) };
    }

    fn __reset_state(plane: *mut bindings::drm_plane, state: Box<Self>) {
        // SAFETY: This would not be called without a plane state to reset, and our data layout is
        // identical to `bindings::drm_plane_state`
        unsafe { bindings::__drm_atomic_helper_plane_reset(plane, Box::into_raw(state).cast()) }
    }

    fn inner(&self) -> &Self::DriverState {
        &self.inner
    }

    fn inner_mut(&mut self) -> &mut Self::DriverState {
        &mut self.inner
    }
}

unsafe impl Zeroable for bindings::drm_plane_state {}

impl<T: DriverPlaneState> Deref for PlaneState<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl<T: DriverPlaneState> DerefMut for PlaneState<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.inner
    }
}

/// A DRM plane state without a known `DriverPlaneState` implementation
///
/// Used in situations where we can't automatically derive the `DriverPlaneState` implementation for
/// a plane state type.
#[repr(transparent)]
pub struct OpaquePlaneState<T: KmsDriver> {
    state: bindings::drm_plane_state,
    _p: PhantomData<T>,
}

impl<T: KmsDriver> AsRawPlaneState for OpaquePlaneState<T> {
    type Driver = T;
    type Plane = OpaquePlane<T>;
}

impl<T: KmsDriver> private::AsRawPlaneState for OpaquePlaneState<T> {
    fn as_raw(&self) -> &bindings::drm_plane_state {
        &self.state
    }

    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_plane_state {
        &mut self.state
    }
}

impl<T: KmsDriver> FromRawPlaneState for OpaquePlaneState<T> {
    unsafe fn from_raw<'a>(ptr: *const bindings::drm_plane_state) -> &'a Self {
        // SAFETY: Our data layout is identical to `ptr`
        unsafe { &*ptr.cast() }
    }

    unsafe fn from_raw_mut<'a>(ptr: *mut bindings::drm_plane_state) -> &'a mut Self {
        // SAFETY: Our data layout is identical to `ptr`
        unsafe { &mut *ptr.cast() }
    }
}

/// A smart-pointer for mutating a DRM plane's state
///
/// Returned by an atomic state mutator (see `AtomicStateMutator` and `AtomicStateComposer`) in
/// order to allow for the mutation of a plane state. In order to uphold rust's data-aliasing
/// rules surrounding mutable references, only one mutable borrow for a plane's state may exist at a
/// time.
///
/// As well, drivers may get a direct mutable reference to their private state data through this
/// type's `DerefMut` implementation.
pub struct BorrowedPlaneState<'a, T: FromRawPlaneState> {
    state: &'a mut T,
    mask: &'a Cell<u32>
}

impl<'a, T: FromRawPlaneState> BorrowedPlaneState<'a, T> {
    pub(super) fn new(
        mutator: &'a AtomicStateMutator<T::Driver>,
        state: NonNull<bindings::drm_plane_state>
    ) -> Option<Self> {
        // SAFETY: `plane` is invariant throughout the lifetime of the atomic state, is
        // initialized by this point, and we're guaranteed it is of type `AsRawPlane` by type
        // invariance
        let plane = unsafe { T::Plane::from_raw((*state.as_ptr()).plane) };
        let plane_mask = plane.mask();
        let borrowed_mask = mutator.borrowed_planes.get();

        if borrowed_mask & plane_mask == 0 {
            mutator.borrowed_planes.set(borrowed_mask | plane_mask);
            Some(Self {
                mask: &mutator.borrowed_planes,
                // SAFETY: We're guaranteed `state` is of `FromRawPlaneState` by type invariance,
                // and we just confirmed by checking `borrowed_planes` that no other mutable borrows
                // have been taken out for `state`
                state: unsafe { T::from_raw_mut(state.as_ptr()) },
            })
        } else {
            None
        }
    }
}

impl<'a, T: FromRawPlaneState> Drop for BorrowedPlaneState<'a, T> {
    fn drop(&mut self) {
        let mask = self.state.plane().mask();
        self.mask.set(self.mask.get() & !mask);
    }
}

impl<'a, T: FromRawPlaneState> AsRawPlaneState for BorrowedPlaneState<'a, T> {
    type Driver = T::Driver;
    type Plane = T::Plane;
}

impl<'a, T: FromRawPlaneState> private::AsRawPlaneState for BorrowedPlaneState<'a, T> {
    fn as_raw(&self) -> &bindings::drm_plane_state {
        self.state.as_raw()
    }

    unsafe fn as_raw_mut(&mut self) -> &mut bindings::drm_plane_state {
        // SAFETY: This function is bound by the same safety contract as `self.inner.as_raw_mut()`
        unsafe { self.state.as_raw_mut() }
    }
}

impl<'a, T: FromRawPlaneState> Sealed for BorrowedPlaneState<'a, T> {}

impl<'a, T: PlaneStateHelper> Deref for BorrowedPlaneState<'a, T> {
    type Target = T::DriverState;

    fn deref(&self) -> &Self::Target {
        self.state.inner()
    }
}

impl<'a, T: PlaneStateHelper> DerefMut for BorrowedPlaneState<'a, T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        self.state.inner_mut()
    }
}

unsafe extern "C" fn plane_destroy_callback<T: DriverPlane>(
    plane: *mut bindings::drm_plane
) {
    // SAFETY: plane contains the only possible entrypoint to this function, so the pointer must be
    // valid
    unsafe { bindings::drm_plane_cleanup(plane) };

    // Reclaim ownership of the reference we took in Plane::<T>::new() so we can drop it
    // SAFETY: We're guaranteed by type invariants this plane is contained within an Box<Plane<T>>
    drop(unsafe { Box::from_raw(plane as *mut Plane<T>) });
}

unsafe extern "C" fn atomic_duplicate_state_callback<T: PlaneStateHelper>(
    plane: *mut bindings::drm_plane
) -> *mut bindings::drm_plane_state {
    // SAFETY: `plane` has to point to a valid instance of drm_plane, since it holds the vtable for
    // this function - which is the only possible entrypoint the caller could have used
    let state = unsafe { (*plane).state };
    if state.is_null() {
        return null_mut();
    }

    // SAFETY: We just checked that the state is non-null, and we're guaranteed this operation is
    // safe via type invariance. As well, we're coercing the mutable reference we receive from
    // `as_raw_mut()` into a pointer which we return to DRM - which is safe.
    unsafe {
        T::from_raw(state)
            .__duplicate_state(plane)
            .map_or(null_mut(), |s| (&mut *Box::into_raw(s)).as_raw_mut())
    }
}

unsafe extern "C" fn atomic_destroy_state_callback<T: PlaneStateHelper>(
    _plane: *mut bindings::drm_plane,
    state: *mut bindings::drm_plane_state
) {
    T::__destroy_state(state);

    // SAFETY: We're guaranteed by type invariance that state is of type T, and since we're being
    // called from the destroy state callback we're guaranteed to be the only ones who could
    // possibly hold a mutable reference to `state`. Furthermore, __destroy_state doesn't deallocate
    // memory - so `state` is still accessible.
    drop(unsafe { Box::from_raw(T::from_raw_mut(state)) });
}

unsafe extern "C" fn plane_reset_callback<T: PlaneStateHelper>(
    plane: *mut bindings::drm_plane,
) {
    // SAFETY: The only entrypoint to this function lives in `plane` so it must be valid
    let state = unsafe { (*plane).state };
    if !state.is_null() {
        // SAFETY: We're guaranteed by type invariance that this planes state is of type T
        unsafe { atomic_destroy_state_callback::<T>(plane, state) }

        // SAFETY: No special requirements here, DRM expects this to be NULL
        unsafe { (*plane).state = null_mut(); }
    }

    // Unfortunately, this is the best we can do at the moment as this FFI callback was mistakenly
    // presumed to be infallible :(
    let new = Box::new(T::default(), GFP_KERNEL).expect("Blame the API, sorry!");

    // DRM takes ownership of the state from here, resets it, and then assigns it to the plane
    T::__reset_state(plane, new);
}

unsafe extern "C" fn atomic_update_callback<T: DriverPlane>(
    plane: *mut bindings::drm_plane,
    state: *mut bindings::drm_atomic_state,
) {
    // SAFETY: We're guaranteed `plane` is of type `Plane<T>` by type invariance, and we're
    // guaranteed it is pointing to a valid state by DRM
    let plane = unsafe { Plane::from_raw(plane) };

    // SAFETY: AtomicStateComposer starts with `bindings::drm_atomic_state` at the beginning of its
    // data layout, and DRM never passes an invalid ptr for `state`
    let state = unsafe { AtomicStateMutator::from_raw(NonNull::new_unchecked(state)) };

    T::atomic_update(plane, &state);
}

unsafe extern "C" fn atomic_check_callback<T: DriverPlane>(
    plane: *mut bindings::drm_plane,
    state: *mut bindings::drm_atomic_state,
) -> i32 {
    // SAFETY: We're guaranteed `plane` is of type `Plane<T>` by type invariance, and we're
    // guaranteed it is pointing to a valid state by DRM
    let plane = unsafe { Plane::from_raw(plane) };

    // SAFETY: DRM never passes an invalid ptr for `state`
    let state = unsafe { LeakyAtomicStateComposer::new(NonNull::new_unchecked(state)) };

    from_result(|| T::atomic_check(plane, &state).map(|_| 0))
}

unsafe extern "C" fn prepare_fb_callback<T: DriverPlane>(
    plane: *mut bindings::drm_plane,
    state: *mut bindings::drm_plane_state,
) -> i32 {
    // SAFETY: We're guaranteed `plane` is of type `Plane<T>` by type invariance, and we're
    // guaranteed it is pointing to a valid state by DRM
    let plane = unsafe { Plane::from_raw(plane) };

    // SAFETY: We're guaranteed `plane` is of type `T::State` by type invariance, and guaranteed
    // it's pointing to a valid state by DRM
    // FIXME: This is a bug - we can't hand the result of from_raw_mut() to a caller, we need a cell
    // for it
    let state = unsafe { T::State::from_raw_mut(state) };

    from_result(|| T::prepare_fb(plane, state).map(|_| 0))
}

unsafe extern "C" fn cleanup_fb_callback<T: DriverPlane>(
    plane: *mut bindings::drm_plane,
    state: *mut bindings::drm_plane_state,
) {
    // SAFETY: We're guaranteed `plane` is of type `Plane<T>` by type invariance, and we're
    // guaranteed it is pointing to a valid state by DRM
    let plane = unsafe { Plane::from_raw(plane) };

    // SAFETY: We're guaranteed `plane` is of type `T::State` by type invariance, and guaranteed
    // it's pointing to a valid state by DRM
    // FIXME: This is a bug - we can't hand the result of from_raw_mut() to a caller
    let state = unsafe { T::State::from_raw_mut(state) };

    T::cleanup_fb(plane, state);
}

#[derive(Copy, Clone, Debug, PartialEq, Eq)]
#[repr(u32)]
pub enum PlaneType {
    OVERLAY = bindings::drm_plane_type_DRM_PLANE_TYPE_OVERLAY,
    PRIMARY = bindings::drm_plane_type_DRM_PLANE_TYPE_PRIMARY,
    CURSOR = bindings::drm_plane_type_DRM_PLANE_TYPE_CURSOR,
}
