//! Rect utilities

use bindings;
use core::{
    ffi::{c_int, c_uint},
    ops::{Deref, DerefMut},
    mem,
};

#[repr(transparent)]
#[derive(Copy, Clone, Default, Eq, PartialEq)]
pub struct Rect(bindings::drm_rect);

impl Rect {
    pub const fn new(x: c_int, y: c_int, w: c_int, h: c_int) -> Self {
        Self(bindings::drm_rect {
            x1: x,
            y1: y,
            x2: x + w,
            y2: y + h,
        })
    }

    pub const fn adjust_size(&self, dw: c_int, dh: c_int) -> Self {
        Self(bindings::drm_rect {
            x1: self.0.x1 - dw >> 1,
            y1: self.0.y1 - dh >> 1,
            x2: self.0.x2 + (dw + 1) >> 1,
            y2: self.0.y2 + (dh + 1) >> 1,
        })
    }

    pub const fn translate(&self, dx: c_int, dy: c_int) -> Self {
        Self(bindings::drm_rect {
            x1: self.0.x1 + dx,
            y1: self.0.y1 + dy,
            x2: self.0.x2 + dx,
            y2: self.0.y2 + dy,
        })
    }

    pub const fn translate_to(&self, x: c_int, y: c_int) -> Self {
        self.translate(x - self.0.x1, y - self.0.y1)
    }

    pub const fn width(&self) -> c_int {
        self.0.x2 - self.0.x1
    }

    pub const fn height(&self) -> c_int {
        self.0.y2 - self.0.y1
    }

    pub const fn visible(&self) -> bool {
        self.width() > 0 && self.height() > 0
    }

    #[inline]
    pub fn rotate(&self, width: c_int, height: c_int, rotation: c_uint) -> Self {
        let mut res = *self;

        // SAFETY: FFI call with no special requirements
        unsafe { bindings::drm_rect_rotate(&mut res.0, width, height, rotation) };

        res
    }
}

impl From<bindings::drm_rect> for Rect {
    fn from(value: bindings::drm_rect) -> Self {
        Self(value)
    }
}

impl AsRef<Rect> for bindings::drm_rect {
    fn as_ref(&self) -> &Rect {
        // SAFETY: Our data layouts are identical
        unsafe { mem::transmute(self) }
    }
}

impl AsMut<Rect> for bindings::drm_rect {
    fn as_mut(&mut self) -> &mut Rect {
        // SAFETY: Our data layouts are identical
        unsafe { mem::transmute(self) }
    }
}
